if True:

    from apsimNGpy.core.apsim import ApsimModel
    from pathlib import Path
    # instantiate the model
    model = ApsimModel(model='Maize', out_path = 'new_maize.apsimx')
    # change the planting density
    model.edit_model(model_type='Models.Manager', model_name='Sow using a variable rule', Population =12)
    # download and replace weather data automatically
    lonlat=  (-93.44, 41.1234)
    model.get_weather_from_web(lonlat = lonlat, start=1981, end=2022, source='daymet')
    # change the start and end date
    model.edit_model(model_type='Models.Clock', model_name='Clock', start="1990-01-01", end="2021-12-31",) # ISO 8601 format (YYYY-MM-DD)
    # run the model
    model.run(report_name="Report")
    # get results
    df = model.results
    # save the model
    filename = str(Path(f'my_edited_maize_model_at{lonlat}.apsimx').resolve())
    model.save(file_name=filename)






from datetime import datetime
from typing import Union
from pydantic import BaseModel, PositiveInt


class User(BaseModel):
    id: Union[int, 'float']
    name: str = 'John Doe'
    signup_ts: datetime | None
    tastes: dict[str, PositiveInt]


external_data = {
    'id': '123',
    'signup_ts': '2019-06-01 12:22',
    'tastes': {
        'wine': 9,
        'cheese': 7,
        'cabbage': '1',
    },
}

user = User(**external_data)

import numpy as np
from pymoo.problems import get_problem
from pymoo.visualization.scatter import Scatter

# The pareto front of a scaled zdt1 problem
pf = get_problem("zdt1").pareto_front()

# The result found by an algorithm
A = pf[::10] * 1.1

# plot the result
Scatter(legend=True).add(pf, label="Pareto-front").add(A, label="Result").show()


import os
import sys
import unittest
from pathlib import Path
from unittest.mock import patch

import pandas as pd
from apsimNGpy.core.config import apsim_version, stamp_name_with_version
from apsimNGpy.core.core import CoreModel, Models
from apsimNGpy.core.model_tools import find_model, validate_model_obj, find_child
from apsimNGpy.settings import logger
# Import the module where CoreModel class is defined
from apsimNGpy.tests.unittests.base_unit_tests import BaseTester
from apsimNGpy.core.base_data import load_default_simulations
from apsimNGpy.core.pythonet_config import is_file_format_modified

IS_NEW_APSIM = is_file_format_modified()

wd = Path.cwd() / "test_core"
wd.mkdir(parents=True, exist_ok=True)
os.chdir(wd)
VERSION = apsim_version()


class TestCoreModel(BaseTester):
    def setUp(self):
        self.user_file_name = stamp_name_with_version(
            os.path.realpath(f"{self._testMethodName}.apsimx"))  # do not delete until tearing up
        self.weather_file_nasa = os.path.realpath('test__met_nasa.met')
        self.weather_file_daymet = os.path.realpath('test_met_daymet.met')
        self.test_user_out_path = stamp_name_with_version(os.path.realpath(f'{self._testMethodName}.apsimx'))
        self.model = 'Maize'
        self.test_ap_sim = CoreModel(self.model, out_path=self.user_file_name)
        self.paths = {self.test_user_out_path,
                      self.weather_file_nasa,
                      self.weather_file_daymet,
                      self.user_file_name}

    def test_out_path(self):
        """
        Test that we can provide an output path for the file name
        @return:
        """
        user_name = stamp_name_with_version(os.path.realpath(f'use_name_ot_test.apsimx'))
        if os.path.exists(user_name):
            os.remove(user_name)
        model = CoreModel('Maize', out_path=user_name)
        in_path = str(model.path)
        # self.assertEqual(in_path, str(user_name))
        self.assertGreater(os.path.getsize(user_name), 0, 'out_path is empty')
        model.clean_up()

    def test_random_out_path(self):
        model = CoreModel(self.model, out_path=None)
        self.assertTrue(os.path.exists(model.path))
        self.paths.add(model.path)
        model.clean_up()

    def test_run(self):

        """ Test the run method's behavior under normal conditions.
         ==========================================================="""

        """There are several ways to check if the run was successful, here I have tried all of them"""
        self.test_ap_sim.run(report_name=['Report'])
        self.assertIsInstance(self.test_ap_sim.results, pd.DataFrame,
                              msg="an error occurred while passing  alist "
                                  "of report names")
        # one more test
        # check if the use pass  report name as str a strict a pandas.core.frame.DataFrame' is returned
        self.test_ap_sim.run(report_name='Report', verbose=False)
        df = self.test_ap_sim.results
        self.assertIsInstance(df, pd.DataFrame)
        # check if it is not empty
        self.assertEqual(df.empty, False, msg='Model was run, but results are empty')
        self.assertTrue(self.test_ap_sim.ran_ok, f'something happened in {repr(self.test_ap_sim)}')
        self.paths.add(self.user_file_name)
        self.test_ap_sim.clean_up(db=True)

    #            self.assertTrue(mock_datastore.Close.called)

    def test_update_mgt(self):
        """ Test the update method's behavior under
        ==========================================="""
        Amount = 23.557777
        script_name = 'Fertilise at sowing'
        mgt_script = {'Name': script_name, 'Amount': Amount},
        self.test_ap_sim.update_mgt(management=mgt_script)
        # extract user infor
        value = self.test_ap_sim.inspect_model_parameters('Manager', simulations='Simulation',
                                                          model_name='Fertilise at sowing', parameters='Amount')
        amountIn = float(value['Amount'])
        self.assertEqual(amountIn, Amount)

    def test_find_simulations(self):
        """ Test find_simulations based on three input None, lists and string
        =====================================================================
        """
        # test None
        sim = 'Simulation'
        MSG = f'test_find_simulations failed to return requested simulation object: {sim}'
        self.assertTrue(self.test_ap_sim.find_simulations(), msg=MSG)
        # test str
        self.assertTrue(self.test_ap_sim.find_simulations(simulations=sim), msg=MSG)
        # test tuple

        self.assertTrue(self.test_ap_sim.find_simulations(simulations=(sim,)), msg=MSG)
        # test list input
        self.assertTrue(self.test_ap_sim.find_simulations(simulations=[sim]), msg=MSG)

    def test_get_simulated_output(self):
        """ Test load_simulated_results
        requires that the models are run first
        ================================================
        """
        self.test_ap_sim.run()
        if not self.test_ap_sim.ran_ok:
            raise unittest.SkipTest(f'skipping test_simulated_results because model did '
                                    f'not run successfully or db was deleted')
        repos = self.test_ap_sim.get_simulated_output(report_names='Report')
        msg = f"expected pd.dataframe but received {type(repos)}"
        self.assertIsInstance(repos, pd.DataFrame, msg=msg)
        self.test_ap_sim.clean_up()

    def test_get_reports(self):
        self.assertIsInstance(self.test_ap_sim.inspect_model('Report'), list)

    def test_inspect_clock(self):
        self.assertIsInstance(self.test_ap_sim.inspect_model('Clock'), list)

    def test_inspect_simulation(self):
        self.assertIsInstance(self.test_ap_sim.inspect_model('Models.Core.Simulation', fullpath=False), list)

    def test_find__simulations(self):
        # get a list of simulations
        sims = self.test_ap_sim.find_simulations()
        self.assertIsInstance(sims, list)
        self.assertTrue(len(sims) > 0)
        sims = self.test_ap_sim.find_simulations(simulations=None)
        # ensures simulations are enclosed in a list
        self.assertIsInstance(sims, list)
        self.assertTrue(len(sims) > 0)
        self.test_ap_sim.clean_up()

    def test_rename(self):
        NEW_NAME = 'NEW_SIM_NAME'
        NEW_SIMs_NAME = 'NEW_SIMULATION_NAME'
        from apsimNGpy.core.base_data import load_default_simulations
        model = load_default_simulations(crop='Maize')
        model.rename_model(model_type="Simulation", old_name='Simulation', new_name=NEW_NAME)
        # check if it has been successfully renamed
        sims = model.inspect_model(model_type='Simulation', fullpath=False)
        assert NEW_NAME in sims, 'rename not successful'
        # The alternative is to use model.inspect_file to see your changes
        model.rename_model(model_type="Simulations", old_name='Simulations', new_name=NEW_SIMs_NAME)
        assert NEW_SIMs_NAME in model.inspect_model(model_type='Simulations',
                                                    fullpath=False), 'renaming simulations was not successful'

    def test_replace_soil_property_values(self):
        parameter = 'Carbon'
        param_values = [2.4, 1.4]
        self.test_ap_sim.replace_soil_property_values(parameter=parameter, param_values=param_values,
                                                      soil_child='Organic', )
        lisT = self.test_ap_sim.inspect_model_parameters(model_type='Organic', simulations='all', model_name='Organic',
                                                         parameters='Carbon')
        lisT = lisT['Carbon'].tolist()
        self.assertIsInstance(lisT, list, msg='expected a list got {}'.format(type(lisT)))
        self.assertTrue(lisT)
        self.assertIsInstance(lisT, list, msg='expected a list got {}'.format(type(lisT)))
        self.assertTrue(lisT)
        # if it was successful
        testP = self.test_ap_sim.inspect_model_parameters(model_type='Organic', model_name='Organic',
                                                          parameters='Carbon')
        testP = testP['Carbon'].tolist()
        self.assertEqual(lisT[:2], param_values,
                         msg=f'replace_soil_property_values was not successful returned {testP}\n got {param_values}')
        self.test_ap_sim.clean_up()

    def test_replace_soil_properties_by_path(self):
        param_values = [1.45, 1.95]
        self.test_ap_sim.edit_model(model_type='Organic', model_name='Organic', Carbon=param_values)

    def test_update_mgt_by_path(self):
        # test when fmt  ='.'
        pathto = '.Simulations.Simulation.Field.Sow using a variable rule'
        self.test_ap_sim.update_mgt_by_path(path=pathto, Population=7.5, fmt='.')
        self.test_ap_sim.run()
        self.assertTrue(self.test_ap_sim.ran_ok, msg='simulation was not ran when fmt was .')
        # test when fmt = '/'
        pathto = '/Simulations/Simulation/Field.Sow using a variable rule'
        self.test_ap_sim.update_mgt_by_path(path=pathto, Population=7.5, fmt='/')
        self.test_ap_sim.run()
        self.assertTrue(self.test_ap_sim.ran_ok, msg='simulation was not ran when fmt was /')

    def test_create_experiment(self):
        """creates a factorial experiment adds a factor, then test if it runs successfully
        =====================================================================================
        """
        if IS_NEW_APSIM:
            pass
            # being tested else where in another file
        else:

            self.test_ap_sim.create_experiment()
            # add factor
            self.test_ap_sim.add_factor(specification="[Fertilise at sowing].Script.Amount = 0 to 200 step 20",
                                        factor_name='Nitrogen')
            self.test_ap_sim.run()
            self.assertTrue(self.test_ap_sim.ran_ok, msg='after adding the experiment and factor running apsim failed')
            self.test_ap_sim.clean_up()

    def test_add_crop_replacements(self):
        self.test_ap_sim.add_crop_replacements(_crop='Maize')
        xp = find_child(self.test_ap_sim.Simulations, child_name='Replacements', child_class='Models.Core.Folder')
        # xp = self.test_ap_sim.Simulations.FindInScope('Replacements')
        if xp:
            rep = True
        else:
            rep = False
        self.assertTrue(rep, msg='replacement was not successful')
        self.test_ap_sim.clean_up()

    def test_replace_soils_values_by_path(self):
        node = '.Simulations.Simulation.Field.Soil.Organic'
        self.test_ap_sim.replace_soils_values_by_path(node_path=node,
                                                      indices=[0], Carbon=1.3)
        v = self.test_ap_sim.get_soil_values_by_path(node, 'Carbon')['Carbon'][0]
        self.assertEqual(v, 1.3)
        self.test_ap_sim.clean_up()

    def test_add_db_table(self):
        self.test_ap_sim.add_db_table(
            variable_spec=['[Clock].Today', '[Soil].Nutrient.TotalC[1]/1000 as SOC1',
                           '[Maize].Grain.Total.Wt*10 as Yield'],
            rename='report4', set_event_names=['[Maize].Harvesting', '[Clock].EndOfYear'])
        # check if report is created
        _reports = self.test_ap_sim.inspect_model('Models.Report()', fullpath=False)

        assert 'report4' in _reports, 'report4 was not found in reports'
        # try running
        self.test_ap_sim.run('report4')

        self.assertTrue(self.test_ap_sim.ran_ok, msg='simulation was not ran after adding the report table report2.')
        self.test_ap_sim.clean_up()

    def test_loading_defaults_with(self):
        model = CoreModel(model='Maize', out_path=Path('ld.apsimx'))
        model.run()
        self.assertTrue(model.ran_ok, 'simulation was not ran after using default within CoreModel class')
        self.paths.add(model.path)
        model.clean_up()

    def test_edit_with_path_som(self):
        model = CoreModel(model='Maize', out_path='path_som.apsimx')
        initial_ratio = 200.0
        model.edit_model_by_path('.Simulations.Simulation.Field.SurfaceOrganicMatter', InitialCNR=initial_ratio,
                                 verbose=False)
        # after inspecting if it was successfully updated
        icnr = model.inspect_model_parameters(model_type='SurfaceOrganicMatter', model_name='SurfaceOrganicMatter',
                                              parameters='InitialCNR')
        self.assertEqual(icnr['InitialCNR'], initial_ratio, msg='InitialCNR was not successfully updated by '
                                                                'edit_model_by_path method')
        self.paths.add(model.path)
        model.clean_up()

    def test_find_model_and_eval_models(self):
        """
        Unit test for the `find_model` and `_eval_model` functions.
        Verifies that:
        - Models can be found correctly by name (case-insensitive).
        - Strings representing full module paths are evaluated correctly.
        """

        # Test finding a model by simple string name
        clok_model = find_model('clock')
        validate_model_obj(clok_model)  # Validate model lookup
        self.assertEqual(clok_model, Models.Clock)  # Ensure it matches the Clock model

        simulation_model = find_model('Simulation')
        validate_model_obj(simulation_model)  # Validate model lookup
        self.assertEqual(simulation_model, Models.Core.Simulation)  # Ensure correct match

        # Test direct evaluation of model paths as strings
        clock = validate_model_obj('Models.Clock')
        self.assertEqual(clock, Models.Clock)  # Confirm evaluation matches Clock model

        simulation = validate_model_obj('Models.Core.Simulation')
        self.assertEqual(simulation, Models.Core.Simulation)  # Confirm evaluation matches Core.Simulation model

        # Test direct evaluation of just model name
        experiment = validate_model_obj('Experiment')
        self.assertEqual(experiment,
                         Models.Factorial.Experiment)  # Confirm evaluation matches Models.Factorial.ExperimentManager model

        factors = validate_model_obj('Factors')
        self.assertEqual(factors, Models.Factorial.Factors)  # Confirm evaluation matches Models.Factorial.Factors model

    def test_add_model_simulation(self):
        if not IS_NEW_APSIM:

            try:
                self.test_ap_sim.add_model('Simulation', adoptive_parent='Simulations', rename='soybean_replaced',
                                           source='Soybean', override=True)

                assert 'soybean_replaced' in self.test_ap_sim.inspect_model('Simulation',
                                                                            fullpath=False), 'testing adding simulations was not successful'
            # clean up
            finally:
                from apsimNGpy.core.core import get_or_check_model

                self.test_ap_sim.remove_model(model_class='Simulation', model_name='soybean_replaced')

    def test_inspect_file(self):
        _inspected = self.test_ap_sim.inspect_file(console=False)
        self.assertTrue(_inspected)

    def test_get_weather_from_web_nasa(self):
        model = load_default_simulations('Maize')
        try:
            model.get_weather_from_web(lonlat=(-93.50456, 42.601247), start=1990, end=2001, source='nasa',
                                       filename=self.weather_file_nasa)
            model.run()
            self.assertTrue(model.ran_ok)
            self.paths.add(model.path)
            model.clean_up()
        except KeyError:
            pass

    def test_get_weather_from_web_daymet(self):
        model = load_default_simulations('Maize')
        model.get_weather_from_web(lonlat=(-93.50456, 42.601247), start=1990, end=2001, source='daymet',
                                   filename=self.weather_file_daymet)
        model.run()
        self.assertTrue(model.ran_ok)
        self.paths.add(model.path)
        model.clean_up()

    def test_edit_model(self):
        self.test_ap_sim.edit_model(model_type='Clock', simulations="Simulation", model_name='Clock',
                                    Start='1900-01-01', End='1990-01-12')
        start = self.test_ap_sim.Start != 'unknown'
        end = self.test_ap_sim.End != 'unknown'
        self.assertTrue(start, 'Simulation start date was not successfully changed')
        self.assertTrue(end, 'Simulation end date was not successfully changed')

    def test_edit_cultivar_edit_model_method(self):
        """
        Test the edit_cultivar requires that we have replacements in place
        @return:
        """
        test_ap_sim = CoreModel('Maize', out_path=self.user_file_name)
        # test_ap_sim.add_crop_replacements(_crop='Maize')

        out_cultivar = 'B_110-e'
        new_juvenile = 289.777729777
        com_path = '[Phenology].Juvenile.Target.FixedValue'
        test_ap_sim.edit_model(model_type='Cultivar', model_name='B_110', new_cultivar_name=out_cultivar,
                               commands=com_path, values=new_juvenile,
                               cultivar_manager='Sow using a variable rule')

        # first we check the current '[Phenology].Juvenile.Target.FixedValue': '211'
        cp = test_ap_sim.inspect_model_parameters(model_type='Cultivar', model_name=out_cultivar)
        self.assertEqual(new_juvenile, float(cp.get(com_path)))
        self.paths.add(test_ap_sim.path)
        test_ap_sim.clean_up()

    def _test_edit_cultivar(self):

        """This is deprecated though"""
        com_path = '[Phenology].Juvenile.Target.FixedValue'

        test_ap_sim = CoreModel("Maize", out_path=self.user_file_name)
        cp = test_ap_sim.inspect_model_parameters(model_type='Cultivar', model_name='B_110')

        new_juvenile = 289.888
        test_ap_sim.edit_cultivar(commands=com_path, values=new_juvenile, CultivarName='B_110')
        # read again
        juvenile_replacements = test_ap_sim.inspect_model_parameters(model_type='Cultivar', model_name='B_110')[
            com_path]
        if juvenile_replacements:
            juvenile_replacements = float(juvenile_replacements)
        self.assertEqual(new_juvenile, juvenile_replacements)
        test_ap_sim.run()
        self.paths.add(test_ap_sim.path)
        test_ap_sim.clean_up()

    def test_inspect_model_parameters(self):

        w_file = self.test_ap_sim.inspect_model_parameters(model_type='Weather', simulations='Simulation',
                                                           model_name='Weather')
        self.assertIsInstance(w_file, str, msg='Inspect weather model parameters failed, when simulation is a str')
        # test a list of simulations
        w_file = self.test_ap_sim.inspect_model_parameters(model_type='Weather', simulations=['Simulation'],
                                                           model_name='Weather')
        self.assertIsInstance(w_file, dict, msg='Inspect weather model parameters failed, when simulation is a list')

    def test_model_info_attributes(self):
        self.assertTrue(self.test_ap_sim.datastore, ' datastore is set to to empty or None')
        self.assertTrue(self.test_ap_sim._DataStore, '_DataStore model is set to to empty or None')
        self.assertTrue(self.test_ap_sim.Datastore, '_DataStore is set to empty or null')

    def test_model_info_types(self):
        self.assertIsInstance(self.test_ap_sim.Datastore, Models.Storage.DataStore,
                              msg='Inspect DataStore, data store was not casted to Models.Storage.DataStore')
        self.assertIsInstance(self.test_ap_sim._DataStore, Models.Storage.DataStore,
                              msg='Inspect DataStore, data store was not casted to Models.Storage.DataStore')

    def test_simulation_container(self):
        self.assertIsInstance(self.test_ap_sim.Simulations, Models.Core.Simulations,
                msg=f'{self.test_ap_sim.Simulations} is expeced to be a Models.Core.Simulations object perhaps casting was not done')

    def tearDown(self):
        self.test_ap_sim.clean_up(db=True)
        for path in self.paths:
            if os.path.exists(path):
                os.remove(path)


if __name__ == '__main__':
    ...
    unittest.main()



import sys
import time

from dataclasses import dataclass, field

from sqlalchemy.testing.util import total_size

# ANSI escape codes for colors
COLOR_CODES = {
    # Standard foreground colors
    'black': '\033[30m',
    'red': '\033[31m',
    'green': '\033[32m',
    'yellow': '\033[33m',
    'blue': '\033[34m',
    'magenta': '\033[35m',
    'cyan': '\033[36m',
    'white': '\033[37m',

    # Bright foreground colors
    'bright_black': '\033[90m',
    'bright_red': '\033[91m',
    'bright_green': '\033[92m',
    'bright_yellow': '\033[93m',
    'bright_blue': '\033[94m',
    'bright_magenta': '\033[95m',
    'bright_cyan': '\033[96m',
    'bright_white': '\033[97m',

    # Background colors
    'bg_black': '\033[40m',
    'bg_red': '\033[41m',
    'bg_green': '\033[42m',
    'bg_yellow': '\033[43m',
    'bg_blue': '\033[44m',
    'bg_magenta': '\033[45m',
    'bg_cyan': '\033[46m',
    'bg_white': '\033[47m',

    # Bright background colors
    'bg_bright_black': '\033[100m',
    'bg_bright_red': '\033[101m',
    'bg_bright_green': '\033[102m',
    'bg_bright_yellow': '\033[103m',
    'bg_bright_blue': '\033[104m',
    'bg_bright_magenta': '\033[105m',
    'bg_bright_cyan': '\033[106m',
    'bg_bright_white': '\033[107m',

    # Reset code
    'reset': '\033[0m'

}


def format_time(seconds):
    millis = int((seconds % 1) * 1000)
    seconds = int(seconds)
    mins, sec = divmod(seconds, 60)
    hrs, mins = divmod(mins, 60)
    return f"{hrs:02d}:{mins:02d}:{sec:02d}.{millis:03d}"

@dataclass
class ProgressBar:
    total: int
    prefix: str = ''
    suffix: str = ''
    length: int = 10
    fill: str = '█'
    color: str = 'green'
    leader_head: str = ' '
    show_time: bool = True
    unit:str = 'iteration'

    iteration: int = field(default=0, init=False)
    start_time: float = field(default_factory=time.perf_counter, init=False)

    def update(self, step: int = 1):
        self.iteration += step
        self._render()
        if self.iteration >= self.total:
            self.close()

    def close(self):
        self.iteration =0
        print()

    def refresh(self, new_total = None):
        """Force a redraw of the current progress bar without changing the iteration."""
        self.total = new_total or self.total
        self._render()

    def _render(self):
        reset = COLOR_CODES['reset']
        color_code = COLOR_CODES.get(self.color.lower(), '')

        percent = f"{100 * (self.iteration / float(self.total)):.1f}"
        filled_length = int(self.length * self.iteration // self.total)
        bar = color_code + self.fill * filled_length + reset + self.leader_head * (self.length - filled_length)
        prefix = COLOR_CODES['red'] + self.prefix + reset
        suffix = COLOR_CODES['red'] + self.suffix + reset
        percent_str = color_code + percent + '%' + reset

        time_info = ''
        tasks =''
        if self.show_time and self.iteration > 0:
            elapsed = time.perf_counter() - self.start_time
            avg_time = elapsed / self.iteration
            formatted_time = format_time(elapsed)
            time_info = f"{color_code} | {avg_time:.2f}s/{self.unit} | Elapsed time: {formatted_time}{reset}"
            tasks = f'{color_code}[{self.iteration}/{self.total}]{reset}'

        sys.stdout.write(f'\r{prefix} |{bar}| {percent_str}| {tasks}| {suffix}{time_info}')
        #sys.stdout.flush()


def print_progress_bar(iteration, total, prefix='', suffix='', length=10, fill='█', color='green', leader_head=' '):
    reset_code  = '\033[0m'
    color_code = COLOR_CODES.get(color.lower(), '')
    percent = f"{100 * (iteration / float(total)):.1f}"
    filled_length = int(length * iteration // total)
    bar = color_code + fill * filled_length + reset_code + leader_head * (length - filled_length)
    prefix = '\033[31m' + prefix + reset_code
    suffix = '\033[31m' + suffix + reset_code
    percent = color_code + percent + reset_code
    p_symbol = color_code + '%' + reset_code

    sys.stdout.write(f'\r{prefix} |{bar}| {percent}{p_symbol} {suffix}' )
    sys.stdout.flush()

    if total==iteration:
        print()

def progress(total, color ='blue'):
    for i in range(total + 1):

        print_progress_bar(i, total, prefix='Progress', suffix='Complete', length=10, color=color, fill='.')
        time.sleep(0.03)  # Simulate work

if __name__ == '__main__':
    # quick tests
  from apsimNGpy.core.apsim import ApsimModel
  bar = ProgressBar(total=50, prefix='Progress', suffix='Complete', color='yellow')
  model = ApsimModel('Maize')
  def fun(x):
      xm = x**3/100000000 * 10000000000000000000000000000000000000000
      time.sleep(1)
      return xm**3 *5+4*2 * 100000000/1000000 * 10000
  for y in range(50):
      model.run()
      bar.update()
  bar.close()
  print(34)





import os
from pathlib import Path
import re

# Define the path to the config file using pathlib for better cross-platform support
DLL_DIR = str(Path(__file__).parent.parent / 'dll')


def start_pythonnet():
    import pythonnet
    try:
        if pythonnet.get_runtime_info() is None:
            return pythonnet.load("coreclr")
    except:
        print("dotnet not found, trying alternate runtime")
        return pythonnet.load()


start_pythonnet()
import clr

clr.AddReference(os.path.join(DLL_DIR, 'CastBridge'))
from CastBridge import CastHelper


# a more pythonic version

def cast_as(model, model_class):
    return CastHelper.CastAs[model_class](model)


simple_rotation_code = """
using Models.Soils;
using System.Collections.Generic;
using APSIM.Shared.Utilities;
using Models.PMF;
using Models.Interfaces;
using Models.Core;
using System;

namespace Models
{
    [Serializable]
    public class Script : Model
    {
        [Link] IClock Clock;
        [Link] ISummary Summary;
        
        [Description("Rotation Sequence (comma separated crop names)")]
        public string[] Crops { get; set; }
        
        // Repeating cycle of input crops. Don't loop on this, loop on Crops
        private IEnumerable<string> cropsEnum()
        {
            var counter = 0;
            while (true)
            {
                yield return Crops[counter];
                ++counter;
                if (counter >= Crops.Length)
                    counter = 0;
            }
        }
        
        private IEnumerator<string> cropCycle;
        
        [Description("Next crop to sow in the simple rotation.")]
        public string NextCrop
        { get => cropCycle.Current; }
        
        [EventSubscribe("StartOfSimulation")]
        private void Init(object sender, EventArgs e)
        {
            // TODO: Error if Crops is empty.
            cropCycle = cropsEnum().GetEnumerator();
            // Set to first value.
            cropCycle.MoveNext();
        }

        [EventSubscribe("PlantEnding")]
        private void OnEndCrop(object sender, EventArgs e)
        {
            // Currently assumes only one crop in the ground, so does not check if the  
            // harvested crop is this script's current crop.
            var old = NextCrop;
            cropCycle.MoveNext();
            var _new = NextCrop;
            Summary.WriteMessage(this, $"Rotation finished with {old}, next up: {_new}.", MessageType.Information);
        }
    }
}
"""

sow_on_fixed_date = """
using APSIM.Shared.Utilities;
using Models.Utilities;
using Models.Soils;
using Models.PMF;
using Models.Core;
using System;
using System.Linq;

namespace Models
{
    [Serializable]
    public class Script : Model
    {
        [Link] Clock Clock;
        [Link] Fertiliser Fertiliser;
        [Link] Summary Summary;
        [Link] Soil Soil;
        
        [Description("Rotation manager name")]
        public Manager RotationManager { get; set; }
        [Separator("Sowing Info")]
        
        [Description("Crop")]
        public IPlant Crop { get; set; }

        [Description("Sowing date (d-mmm)")]
        public string SowDate { get; set; }

        [Display(Type = DisplayType.CultivarName)]
        [Description("Cultivar to be sown")]
        public string CultivarName { get; set; }

        [Description("Sowing depth (mm)")]
        public double SowingDepth { get; set; }

        [Description("Row spacing (mm)")]
        public double RowSpacing { get; set; }

        [Description("Plant population (/m2)")]
        public double Population { get; set; }

        [EventSubscribe("DoManagement")]
        private void OnDoManagement(object sender, EventArgs e)
        {
            if (DateUtilities.WithinDates(SowDate, Clock.Today, SowDate))
            {
                Crop.Sow(population: Population, cultivar: CultivarName, depth: SowingDepth, rowSpacing: RowSpacing);    
            }
        }
    }
}
"""

sow_using_variable_rule = """
using Models.Interfaces;
using APSIM.Shared.Utilities;
using Models.Utilities;
using Models.Soils;
using Models.PMF;
using Models.Core;
using System;
using Models.Climate;

namespace Models
{
    [Serializable]
    public class Script : Model
    {
        [Link] private Clock Clock;
        [Link] private Summary Summary;
        private Accumulator accumulatedRain;
        [Link]
        private ISoilWater waterBalance;
        
        [Description("Simple Rotation Manager")]
        public Manager Rotations { get; set; }
        
        [Separator("Sowing Info")]
        
        [Description("Crop")]
        public IPlant Crop { get; set; }

        [Description("Start of sowing window (d-mmm)")]
        public string StartDate { get; set; }

        [Description("End of sowing window (d-mmm)")]
        public string EndDate { get; set; }

        [Description("Minimum extractable soil water for sowing (mm)")]
        public double MinESW { get; set; }

        [Description("Accumulated rainfall required for sowing (mm)")]
        public double MinRain { get; set; }

        [Description("Duration of rainfall accumulation (d)")]
        public int RainDays { get; set; }

        [Display(Type = DisplayType.CultivarName)]
        [Description("Cultivar to be sown")]
        public string CultivarName { get; set; }

        [Description("Sowing depth (mm)")]
        public double SowingDepth { get; set; }

        [Description("Row spacing (mm)")]
        public double RowSpacing { get; set; }

        [Description("Plant population (/m2)")]
        public double Population { get; set; }
        
        [Description("Must sow by end date?")]
        public bool MustSow { get; set; }
        
        
        [EventSubscribe("StartOfSimulation")]
        private void OnSimulationCommencing(object sender, EventArgs e)
        {
            accumulatedRain = new Accumulator(this, "[Weather].Rain", RainDays);
        }

        [EventSubscribe("DoManagement")]
        private void OnDoManagement(object sender, EventArgs e)
        {
            accumulatedRain.Update();
            if (!Crop.IsAlive &&
                   DateUtilities.WithinDates(StartDate, Clock.Today, EndDate))
               {
                   var rainSum = accumulatedRain.Sum;
                   var eswSum  = MathUtilities.Sum(waterBalance.ESW);
                   if (rainSum > MinRain && eswSum > MinESW ||
                       DateUtilities.DatesEqual(EndDate, Clock.Today) && MustSow)
                   {
                       var thisCrop = Crop.Name.ToLower();
                       var nextCrop = (string)Rotations?.FindByPath("Script.NextCrop")?.Value;
                       if (thisCrop == (nextCrop?.ToLower() ?? thisCrop))
                           Crop.Sow(population: Population,
                                 cultivar: CultivarName,
                                 depth: SowingDepth,
                                 rowSpacing: RowSpacing);
                   }
               }
        }
    }
}
"""

harvest = """using APSIM.Shared.Utilities;
        using Models.Utilities;
        using Models.Soils;
        using Models.PMF;
        using Models.Core;
        using System;
        using System.Linq;
        
        namespace Models
        {
            [Serializable] 
            public class Script : Model
            {
                [Description("Crop")]
                public IPlant Crop { get; set; }
                
                [EventSubscribe("DoManagement")]
                private void OnDoManagement(object sender, EventArgs e)
                {
                    if (Crop.IsReadyForHarvesting)
                    {
                        Crop.Harvest();
                        Crop.EndCrop();
                    }
                }
            }
        }
       """

fertilizer_at_sow = """
    using Models.Soils;
using System;
using System.Linq;
using Models.Core;
using Models.PMF;
namespace Models
{
    [Serializable]
    public class Script : Model
    {
        [Link] Clock Clock;
        [Link] Fertiliser Fertiliser;
        
        [Description("Crop to be fertilised")]
        public IPlant Crop { get; set; }

        [Description("Type of fertiliser to apply? ")] 
        [Display(Type = DisplayType.FertiliserType)]public string FertiliserType { get; set; }
    
        [Description("Amount of fertiliser to be applied (kg/ha)")]
        public double Amount { get; set; }
        
        [Description("Fertilise depth (mm)")]
        public double FertiliseDepth { get; set; }
        
        [EventSubscribe("Sowing")]
        private void OnSowing(object sender, EventArgs e)
        {
            Model crop = sender as Model;
            if (Crop != null && crop.Name.ToLower() == (Crop as IModel).Name.ToLower())
                Fertiliser.Apply(amount: Amount, type: FertiliserType, depth: FertiliseDepth);
        }
    }
}
"""
new_property = '\n        public int RotationIndex { get; set; }\n'

import re


def update_manager_code(old_code, description, typeof, var_name):
    getter = '{ get; set; }'
    new_variable = [
        f'        [Description("{description}")]',
        f'        public {typeof} {var_name}  {getter}',
        ''
    ]

    lines = old_code.splitlines()

    for i in range(1, len(lines)):
        # Match line with [Description(...)]
        if re.match(r'\s*\[Description\(".*?"\)\]', lines[i - 1]) and \
                re.match(r'\s*public\s+[\w<>\[\]]+\s+\w+\s*\{\s*get;\s*set;\s*\}', lines[i]):
            insert_at = i + 2  # skip one line after the property
            lines[insert_at:insert_at] = new_variable
            break  # insert only once

    # Join and print or store
    updated_code = '\n'.join(lines)
    return updated_code

# simple_rotation_code = update_manager_code(simple_rotation_code, 'rotation', typeof='String')



import os.path
import unittest
from pathlib import Path
import pandas as pd
import warnings
from apsimNGpy.core_utils.utils import timer
from apsimNGpy.core.apsim import ApsimModel
from apsimNGpy.core.model_tools import validate_model_obj, ModelTools, add_as_simulation, find_model, \
    get_or_check_model, detect_sowing_managers, find_child, find_all_in_scope, add_replacement_folder, \
    add_model_as_a_replacement, get_or_check_model
from apsimNGpy.core.pythonet_config import is_file_format_modified

IS_NEW_APSIM = is_file_format_modified()


class TestModelTools(unittest.TestCase):
    def setUp(self) -> None:
        self.apsim_path = Path(f"{self._testMethodName}.apsimx")

    def test_add_a_simulation(self):
        # if IS_NEW_APSIM:
        model = ApsimModel('Maize')
        ans = [add_as_simulation(model, i, i) for i in ['Soybean']]
        sims = len(ans)

        t_add = sims >= 1
        self.assertTrue(t_add)

    def test_validate_model_obj(self):
        model = 'Report'
        obj = validate_model_obj(model)
        self.assertTrue(hasattr(obj, '__class__'))

    def test_find_model(self):
        result = find_model('Clock')
        self.assertIsNotNone(result)

    def test_get_or_check_model_get(self):
        model = ApsimModel('Maize', out_path=self.apsim_path)
        model_class = find_model("Clock")
        result = get_or_check_model(model.Simulations, model_class, 'Clock', 'get')
        self.assertTrue(result)

    def test_get_or_check_model_delete(self):
        model = ApsimModel('Maize', out_path=self.apsim_path)
        results = get_or_check_model(model.Simulations, 'Models.Core.Simulation', 'Simulation', action='delete')
        self.assertIsNone(results)
        # ensure model is deleted
        child = find_child(model.Simulations, 'Models.Core.Simulations', child_name='Simulation')
        self.assertIsNone(child, 'child found implies was not deleted successfully')

    def test_get_or_check_model_check(self):
        model = ApsimModel('Maize', out_path=self.apsim_path)
        results = get_or_check_model(model.Simulations, 'Models.Core.Simulation', 'Simulation', action='check')
        res = results == True
        self.assertTrue(res, msg='checking is not doing the desired thing')

    def test_detect_sow_manager(self):
        model = ApsimModel('Maize', out_path=self.apsim_path)
        sowing_manager = detect_sowing_managers(model)
        self.assertIsNotNone(sowing_manager)

    def test_detect_sow_manager_when_none(self):
        model = ApsimModel('Maize', out_path=self.apsim_path)
        sowing_manager = detect_sowing_managers(model)
        ModelTools.DELETE(sowing_manager)
        model.save()
        sowing_manager = detect_sowing_managers(model)
        self.assertIsNone(sowing_manager)

    def test_get_or_check_model_delete1(self):
        model = ApsimModel('Maize', out_path=self.apsim_path)
        model_class = find_model("Clock")
        result = get_or_check_model(model.Simulations, model_class, 'Clock', 'delete')
        self.assertIsNone(result)

    def test_get_or_check_model_check2(self):
        model = ApsimModel('Maize', out_path=self.apsim_path)
        model_class = find_model("Clock")
        result = get_or_check_model(model.Simulations, model_class, 'Clock', 'check')
        self.assertTrue(result)

    def test_find_child(self):
        """test for find_child for a clock"""
        model = ApsimModel('Maize', out_path=self.apsim_path)
        clock = find_child(model.Simulations, 'Models.Clock', 'Clock')
        self.assertIsNotNone(clock, msg='failed to find clock child from simulations')

    def test_find_child_cultivar(self):
        """test find_child of Cultivar"""
        model = ApsimModel('Maize', out_path=self.apsim_path)
        cultivar = find_child(parent=model.Simulations, child_class='Models.PMF.Cultivar', child_name='B_110')
        self.assertIsNotNone(cultivar, msg='failed to find cultivar child from simulations')

    def test_find_child_simulations(self):
        """test finding child simulations"""
        model = ApsimModel('Maize', out_path=self.apsim_path)
        simulation = find_child(parent=model.Simulations, child_class='Models.Core.Simulation', child_name='Simulation')
        self.assertIsNotNone(simulation, msg='failed to find simulation child from simulations')

    def test_find_child_weather(self):
        """test finding child simulations"""
        model = ApsimModel('Maize', out_path=self.apsim_path)
        weather = find_child(parent=model.Simulations, child_class='Models.Climate.Weather', child_name='Weather')
        self.assertIsNotNone(weather, msg='failed to find weather  child from simulations')

    def test_find_all_in_scope_managers(self):
        """test finding all managers"""
        model = ApsimModel('Maize', out_path=self.apsim_path)
        manager = find_all_in_scope(parent=model.Simulations, child_class='Models.Manager')
        # ensure something is found
        self.assertIsNotNone(manager, msg='failed to find weather  child from simulations')

    def test_find_all_in_scope_clock(self):
        """test finding all managers"""
        model = ApsimModel('Maize', out_path=self.apsim_path)
        clock = find_all_in_scope(parent=model.Simulations, child_class='Models.Clock')
        # ensure something is found
        self.assertIsNotNone(clock, msg='failed to find all clock children from simulations')

    def test_find_all_in_scope_physical(self):
        """test finding all managers"""
        model = ApsimModel('Maize', out_path=self.apsim_path)
        physical = find_all_in_scope(parent=model.Simulations, child_class='Models.Soils.Physical')
        # ensure something is found
        self.assertIsNotNone(physical, msg='failed to find all clock children from simulations')

    def test_find_all_in_scope_simulations(self):
        """test finding all managers"""
        model = ApsimModel('Maize', out_path=self.apsim_path)
        simulations = find_all_in_scope(parent=model.Simulations, child_class='Models.Core.Simulations')
        # ensure something is found
        self.assertIsNotNone(simulations, msg='failed to find all manager children from simulations')

    def test_find_all_in_scope_return_emptyforNone(self):
        """test finding all managers"""
        model = ApsimModel('Maize', out_path=self.apsim_path)
        experiments = find_all_in_scope(parent=model.Simulations, child_class='Models.Factorial.Experiment')
        # ensure that it is empty because experiment does not exist at this node
        test = experiments == []
        self.assertTrue(test, msg='failed to return empty for no existing experiments class models')

    def test_add_folder(self):
        'Ensure that replacement folder is added'
        model = ApsimModel('Maize', out=os.path.realpath('replacement2.apsimx'))
        add_replacement_folder(model.Simulations)
        folder = find_child(model.Simulations, 'Models.Core.Folder', "Replacements")
        self.assertTrue(folder, 'replacement folder not found')

    def test_add_crop_replacement(self):
        model = ApsimModel('Maize', out_path=self.apsim_path)
        add_model_as_a_replacement(model.Simulations, 'Models.PMF.Plant', 'Maize')
        folder = find_child(model.Simulations, 'Models.Core.Folder', "Replacements")
        self.assertTrue(folder, 'replacement folder not found')

    def test_add_crop_replacement_and_run(self):
        """ensures that model runs after replacement"""
        model = ApsimModel('Maize', out_path=self.apsim_path)
        add_model_as_a_replacement(model.Simulations, 'Models.PMF.Plant', 'Maize')
        folder = find_child(model.Simulations, 'Models.Core.Folder', "Replacements")
        model.run()
        self.assertTrue(model.ran_ok, 'after replacement, model did not run properly')

    def test_add_clock_replacement(self):
        model = ApsimModel('Maize', out_path=self.apsim_path)
        add_model_as_a_replacement(model.Simulations, 'Models.PMF.Plant', 'Maize')
        folder = find_child(model.Simulations, 'Models.Core.Folder', "Replacements")
        self.assertTrue(folder, 'replacement folder not found')

    def test_add_what_happens_to_repeat_reps(self):
        model = ApsimModel('Maize', out=os.path.realpath('replacement4.apsimx'))
        add_model_as_a_replacement(model.Simulations, 'Models.Clock', 'Clock')
        add_model_as_a_replacement(model.Simulations, 'Models.Clock', 'Clock')
        folder = find_child(model.Simulations, 'Models.Core.Folder', "Replacements")
        self.assertTrue(folder, 'replacement folder not found')

    def test_add_what_happens_to_repeat_reps_with_run(self):
        model = ApsimModel('Maize', out_path=self.apsim_path)
        add_model_as_a_replacement(model.Simulations, 'Models.Clock', 'Clock')
        add_model_as_a_replacement(model.Simulations, 'Models.Clock', 'Clock')
        folder = find_child(model.Simulations, 'Models.Core.Folder', "Replacements")
        self.assertTrue(folder, 'replacement folder not found')
        model.run()
        self.assertTrue(model.ran_ok)

    def tearDown(self):
        self.apsim_path.unlink(missing_ok=True)


# Optional utility function
@timer
def datatable_to_dataframe(dt) -> pd.DataFrame:
    columns = [col.ColumnName for col in dt.Columns]
    data = []
    for row in dt.Rows:
        data.append([row[col] for col in columns])
    return pd.DataFrame(data, columns=columns)


if __name__ == '__main__':
    unittest.main()



"""
This allows for mixed variable optimization by encoding the categorical _variables
"""
from pathlib import Path

import numpy as np
from abc import abstractmethod
from apsimNGpy.core_utils.utils import timer
from scipy.optimize import minimize, differential_evolution
import subprocess

from apsimNGpy.optimizer._one_obj import AbstractProblem, SIMULATIONS, ContVarProblem, cache, VarDesc, SING_OBJ_MIXED_VAR
from apsimNGpy.core_utils.progbar import ProgressBar
try:
    import wrapdisc
except ModuleNotFoundError as mnf:
    print('installing wrapdisc package')
    package_name = "wrapdisc"
    command = ["pip", "install", package_name]
    # Running the command
    subprocess.run(command, check=True)

import operator
from typing import Any

import scipy.optimize

from wrapdisc import Objective
from wrapdisc.var import ChoiceVar, GridVar, QrandintVar, QuniformVar, RandintVar, UniformVar


@cache
def _variable_type(type_name: str) -> str:
    variable_types = {
        'choice': ChoiceVar,
        'grid': GridVar,
        'qrandint': QrandintVar,
        'quniform': QuniformVar,
        'randint': RandintVar,
        'uniform': UniformVar
    }
    try:
        return variable_types[type_name.lower()]
    except KeyError:
        raise ValueError(f"Invalid type '{type_name}'. Use one of: {', '.join(var_map)}")


class MixedVarProblem(ContVarProblem):
    """
           Defines an optimization problem for continuous variables in APSIM simulations.

           This class enables the user to configure and solve optimization problems involving continuous
           control variables in APSIM models. It provides methods for setting up control variables,
           applying bounds and starting values, inserting variable values into APSIM model configurations,
           and running optimization routines using local solvers or differential evolution.

           Inherits from:
               ``ContinuousVariableProblem``

           Parameters:
               ``model (str):`` The name or path of the APSIM template file.
               .
               ``simulation (str or list, optional)``: The name(s) of the APSIM simulation(s) to target.
                                                   Defaults to all simulations.

               ``decision_vars`` (list, optional): A list of VarDesc instances defining variable metadata.

               ``labels (list, optional)``: Variable labels for display and results tracking.

               ``cache_size (int):`` Maximum number of results to store in the evaluation cache.

           Attributes:
               ``model (str):`` The APSIM model template file name.
               ``simulation (str):`` Target simulation(s).
               ``decision_vars (list):`` Defined control variables.
               ``decision_vars (list):`` List of VarDesc instances for optimization.
               ``labels (list): Labels`` for variables.
               ``pbar (tqdm):`` Progress bar instance.
               ```cache (bool):`` Whether to cache evaluation results.
               ```cache_size (int):`` Size of the local cache.

           Methods:
               ``add_control(...):`` Add a new control variable to the optimization problem.
               ``bounds:`` Return the bounds for all control variables as a tuple.
               ``starting_values():`` Return the initial values for all control variables.
               ``minimize_with_local_solver(...):`` Optimize using `scipy.optimize.minimize`.
               ``optimize_with_differential_evolution(...):`` Optimize using `scipy.optimize.differential_evolution`.
               ``_open_pbar(labels, maxiter):`` Open a progress bar.
               ``_close_pbar():`` Close the progress bar.

           Example:
               >>> class Problem(ContVarProblem):
               ...     def evaluate(self, x):
               ...         return -self.run(verbose=False).results.Yield.mean()

               >>> problem = Problem(model="Maize", simulation="Sim")
               >>> problem.add_control("Manager", "Sow using a rule", "Population", int, 5, bounds=[2, 15])
               >>> result = problem.minimize_with_local_solver(method='Powell')
               >>> print(result.x_vars)
           """

    def __init__(self, apsim_model: 'ApsimNGpy.Core.Model', max_cache_size=400, objectives=None, decision_vars=None):

        # Initialize parent classes explicitly
        AbstractProblem.__init__(self, apsim_model, max_cache_size, objectives, decision_vars)
        self.model = apsim_model
        self.decision_vars = decision_vars or []
        self.objectives = objectives

        self.cache = True
        self.cache_size = max_cache_size
        self.pbar = None
        self.counter = 0
        self.maxiter = 0

    def wrap_vars(self):
        wrapped_variables = []

        for var in self.decision_vars:
            var_type = var['v_type']
            bounds = var.get('bounds')
            categories = var.get('categories')
            q = var.get('q', 1)

            # Map variable type to wrapdisc class
            type_map = {
                'int': RandintVar,
                'float': UniformVar,
                'uniform': UniformVar,
                'choice': ChoiceVar,
                'grid': ChoiceVar,  # Alias
                'categorical': ChoiceVar,
                'qrandint': QrandintVar,
                'quniform': QuniformVar
            }

            if var_type not in type_map:
                raise ValueError(f"Unsupported variable type: {var_type}")

            wrap_class = type_map[var_type.lower()]

            if wrap_class in {RandintVar, UniformVar}:
                if bounds is None:
                    raise ValueError(f"Bounds must be provided for variable: {var}")
                if wrap_class is RandintVar:
                    wrapped = RandintVar(lower=bounds[0], upper=bounds[1])
                else:
                    wrapped = UniformVar(lower=bounds[0], upper=bounds[1])
            elif wrap_class is ChoiceVar:
                if categories is None:
                    raise ValueError(f"'values' must be provided for categorical or grid variable: {var}")
                wrapped = ChoiceVar(categories=categories)
            elif wrap_class is QrandintVar:
                wrapped = QrandintVar(lower=bounds[0], upper=bounds[1], q=q)
            elif wrap_class is QuniformVar:
                wrapped = QrandintVar(lower=bounds[0], upper=bounds[1], q=q)

            else:
                raise NotImplementedError(f"Handler not implemented for variable type: {var_type}")

            wrapped_variables.append(wrapped)
            # this is the last time we need them


        return wrapped_variables

    @property
    def _variables(self):

        return self.wrap_vars()

    def _set_objective_function(self, x):
        self._insert_controls(x)
        SCORE = self.evaluate_objectives()
        return SCORE

    def  optimization_type(self):
         return SING_OBJ_MIXED_VAR

    @abstractmethod
    def minimize_with_alocal_solver(self, **kwargs):
        'To be implimneted in sub class'
        pass

    @abstractmethod
    def minimize_with_de(self):
        """
        To be implimented in sub-class
        """
if __name__ == "__main__":
    maize_model = "Maize"

    obs = [
        7000.0,
        5000.505,
        1000.047,
        3504.000,
        7820.075,
        7000.517,
        3587.101,
        4000.152,
        8379.435,
        4000.301
    ]
    class Problem(MixedVarProblem):
        def __init__(self, apsim_model: 'ApsimNGpy.Core.Model', obs, max_cache_size=400, objectives=None,
                     decision_vars=None):
            super().__init__(apsim_model=apsim_model, max_cache_size=max_cache_size, objectives=objectives,
                             decision_vars=decision_vars)
            self.obs = obs

        def evaluate_objectives(self, **kwargs):
            # set up everything you need here
            predicted = self.apsim_model.run(verbose=False).results.Yield
            ans = self.rmse(self.obs, predicted)

            return -predicted.mean()


    def maximize_yield(df):
        return -df.Yield.mean()


    from apsimNGpy.core.cal import OptimizationBase

    maize_model = OptimizationBase('Maize')
    problem = Problem(maize_model, obs, objectives=maximize_yield)
    problem.add_control(
        **{'path': '.Simulations.Simulation.Field.Fertilise at sowing', 'Amount': "?", "bounds": None,
           "v_type": "choice"}, start_value=100,  categories= [100, 200, 250, 300])
    problem.add_control(
        **{'path': '.Simulations.Simulation.Field.Sow using a variable rule', 'Population': "?", 'v_type': 'qrandint',
           'bounds': [4, 14]}, start_value=8, q=2, )
    # de_res = problem.optimize_with_differential_evolution()
    res = problem.minimize_with_alocal_solver(method='Powell', )
    de = problem.minimize_with_differential_evolution(popsize=20, polish =True)



import gc
import os
import re
import sys

from apsimNGpy.core.apsim import ApsimModel
from collections import OrderedDict
from apsimNGpy.core.model_tools import ModelTools, Models
from apsimNGpy.core.cs_resources import CastHelper
from apsimNGpy.core.pythonet_config import is_file_format_modified
from apsimNGpy.core.config import APSIM_VERSION_NO, BASE_RELEASE_NO
from apsimNGpy.core.model_loader import to_json_string
if is_file_format_modified():
    import APSIM.Core as NodeUtils
    import System
    structure = Models.Core.ApsimFile.Structure
else:
    from apsimNGpy.core.config import apsim_version

    raise ValueError(f"The experiment module is not supported for this type of {apsim_version()} ")


class ExperimentManager(ApsimModel):
    def __init__(self, model, out_path=None, out=None):
        super().__init__(model=model, out_path=out_path, out=out)
        self.parent_factor = None
        self.experiment_node = None
        self.factorial_node = None
        self.permutation_node = None
        self.factors = OrderedDict()
        self.specs = OrderedDict()
        self.counter = 0
        self.sims = self.simulations
        self.init = False

    def init_experiment(self, permutation=True):

        """
            Initializes the factorial experiment structure inside the APSIM file.

            Args:
                permutation (bool): If True, enables permutation mode; otherwise, uses standard factor crossing.

            Side Effects:
                Replaces any existing ExperimentManager node with a new configuration.
                Clones the base simulation and adds it under the experiment.
            """
        self.permutation = permutation

        def exp_refresher(mode):
            sim = mode.simulations[0]
            base = ModelTools.CLONER(sim)
            for simx in mode.simulations:  # it does not matter how many experiments exist, we need only one
                ModelTools.DELETE(simx)
            # replace before delete

            try:
                mode.simulations[0] = base
                base = mode.simulations[0]
            except IndexError:
                pass
            experiment = Models.Factorial.Experiment()
            self.experiment_node = experiment
            factor = Models.Factorial.Factors()
            self.factorial_node = factor
            if self.permutation:
                perm_node = Models.Factorial.Permutation()
                self.permutation_node = perm_node
                factor.AddChild(perm_node)
            experiment.AddChild(factor)
            experiment.AddChild(base)
            experi = ModelTools.find_child_of_class(mode.Simulations, Models.Factorial.Experiment)

            if experi:
                ModelTools.DELETE(experi)
            mode.model_info.Node.AddChild(experiment)
            sim_final = CastHelper.CastAs[Models.Core.Simulations](mode.model_info.Node)

            if APSIM_VERSION_NO > BASE_RELEASE_NO:

                simx = ModelTools.find_all_in_scope(sim_final, Models.Core.Simulation)
                simy= [ModelTools.CLONER(i) for i in simx]

                simx = [CastHelper.CastAs[Models.Core.Simulations](i.Node) for i in simy]

                ...

            else:
                simx = list(sim_final.FindAllDescendants[Models.Core.Simulation]())

                if not mode.simulations:
                    mode.simulations.extend(simx)
            # mode.save()

        def refresher():
            siM = NodeUtils.Node.Create(Models.Core.Simulations())
            siM.AddChild(Models.Storage.DataStore())
            # create experiment
            experiment = Models.Factorial.Experiment()
            self.experiment_node = experiment
            factor = Models.Factorial.Factors()
            self.factorial_node = factor
            # branch if it is a permutation experiment
            if self.permutation:
                perm_node = Models.Factorial.Permutation()
                self.permutation_node = perm_node
                factor.AddChild(perm_node)
            experiment.AddChild(factor)
            # add simulation before experiment to the simulation tree
            sim = self.simulations[0]
            siM.AddChild(experiment)
            experiment.AddChild(sim)
            siM = CastHelper.CastAs[Models.Core.Simulations](siM.Model)
            #siM.Write(self.path)
            self.Simulations = siM

            # add experiment
        if APSIM_VERSION_NO > BASE_RELEASE_NO:
            refresher()
        else:
            exp_refresher(self)
        self.init = True

        c= gc.collect()


    def add_factor(self, specification: str, factor_name: str = None, **kwargs):
        """
           Adds a new factor to the experiment based on an APSIM script specification.

           Args:
               specification (str): A script-like APSIM expression that defines the parameter variation.
               factor_name (str, optional): A unique name for the factor; auto-generated if not provided.
               **kwargs: Optional metadata or configuration (not yet used internally).

           Raises:
               ValueError: If a Script-based specification references a non-existent or unlinked manager script.

           Side Effects:
               Inserts the factor into the appropriate parent node (Permutation or Factors).
               If a factor at the same index already exists, it is safely deleted before inserting the new one.
           """
        if not self.init:
            raise ValueError("Please initialize the experiment first by calling: self.init_experiment method")
        # Auto-generate factor name from specification if not provided
        if factor_name is None:
            factor_name = specification.split("=")[0].strip().split(".")[-1]

        # If it's a Script-based specification, validate linkage
        if 'Script' in specification:
            matches = re.findall(r"\[(.*?)\]", specification)
            if matches:
                manager_names = set(self.inspect_model('Models.Manager', fullpath=False))
                linked = set(matches) & manager_names
                if not linked:
                    if matches:
                        matches = ', '.join(matches)
                    raise ValueError(f'Specification{specification} has no linked script `{matches}` in the model')

        # Record factor info
        self.factors[factor_name] = (specification, kwargs)

        # Choose parent node and parent class
        parent_factor = self.permutation_node if self.permutation else self.factorial_node
        parent_class = Models.Factorial.Permutation if self.permutation else Models.Factorial.Factors
        if APSIM_VERSION_NO > BASE_RELEASE_NO:
            parent_factor = ModelTools.find_child_of_class(self.Simulations, parent_class)

        new_factor = Models.Factorial.Factor()
        new_factor.Name = factor_name
        new_factor.set_Specification(specification)

        self.specs[factor_name] = new_factor

        # Maintain counter and avoid index error
        index = len(self.specs) - 1
        try:
            # Try to remove existing child at index before inserting
            if 0 <= index < len(parent_factor.Children):
                old_child = parent_factor.Children[index]
                if old_child is not None:
                    ...
                    parent_factor.Children.Remove(old_child)
                    # NodeUtils.Node.RemoveChild(old_child)
                #  ModelTools.DELETE(old_child)

        except System.ArgumentOutOfRangeException:
            pass

        # Insert a new factor
        parent_factor.Children.Add(new_factor)
        self.parent_factor = parent_factor

    @property
    def n_factors(self):
        """
            Returns:
                int: The total number of active factor specifications currently added to the experiment.
            """
        return len(self.specs)

    def finalize(self):
        """"
        Finalizes the experiment setup by re-creating the internal APSIM factor nodes from specs.

        This method is designed as a guard against unintended modifications and ensures that all
        factor definitions are fully resolved and written before saving.

        Side Effects:
            Clears existing children from the parent factor node.
            Re-creates and attaches each factor as a new node.
            Triggers model saving.
    """
        self.parent_factor.Children.Clear()
        for name, spec in self.specs.items():
            node = NodeUtils.Node.Create(spec, parent=self.parent_factor)
            self.parent_factor.AddChild(node.Model)
        self.save()


if __name__ == '__main__':
    exp = ExperimentManager('Soybean', out_path='exp.apsimx')
    exp.init_experiment(permutation=True)
    exp.add_factor("[Fertilise at sowing].Script.Amount = 0 to 200 step 20")
    exp.add_factor("[Fertilise at sowing].Script.FertiliserType= DAP,NO3N")
    exp.add_factor(specification="[Sow using a variable rule].Script.RowSpacing = 100, 450, 700",
                   factor_name='Population')
    exp.add_factor(specification="[Sow using a variable rule].Script.RowSpacing = 100, 450, 700",
                   factor_name='Population')
    # exp.add_factor(specification="[Sow using a variable rule].Script.RowSpacing = 100, 450, 700",
    #                factor_name='Population')
    # exp.finalize()
    # exp.preview_simulation()



import numpy as np

from pymoo.indicators.hv import HV


def edit_runner(model_runner, decision_specs: dict, x_values, verbose=False):
    """
    Applies a new value to the APSIM model by editing the correct parameter in the model_runner.

    Parameters
    ----------
    model_runner : apsimNGpy.core.cal.OptimizationBase
        The APSIM runner object.
    decision_specs : dict
        A dictionary containing keys: 'path', 'v_type', 'bounds', and 'kwargs'.
        'kwargs' is expected to contain exactly one parameter to be optimized, marked as '?' or ''.
    x_values : float or int
        The value to apply to the unspecified parameter.
    verbose : bool
        If True, print what is being edited.
    """
    path = decision_specs['path']
    kwargs = decision_specs.get('kwargs', decision_specs).copy()
    kwargs.pop('path', None)
    kwargs.pop('v_type', None)
    kwargs.pop('bounds', None)
    kwargs.pop('q', None)
    kwargs.pop('values', None)
    kwargs.pop('start_value', None)

    # Replace the placeholder (e.g., '?' or '') with the optimization value
    fill_keys = [k for k, v in kwargs.items() if v in ('?', '',)]
    if len(fill_keys) > 1:
        raise ValueError("Exactly one parameter should be marked with '?' or '' for optimization.")

    param_key = fill_keys[0]
    kwargs[param_key] = x_values

    if verbose:
        print(f"Editing {path}: setting {param_key} = {x_values}")

    model_runner.edit_model_by_path(path=path, **kwargs)


def compute_hyper_volume(F, reference_point=None, normalize=False, normalization_bounds=None):
    """
    Compute the hypervolume of a Pareto front.

    Parameters
    ----------
    F : np.ndarray
        A 2D array of shape (n_points, n_objectives), typically `result.F` from pymoo.

    reference_point : list or array, optional
        Reference point for hypervolume. If None:
        - Raw: uses max values in F + 10%
        - Normalized: uses [1.0, 1.0, ..., 1.0]

    normalize : bool, default=False
        Whether to normalize the objectives to [0, 1].

    normalization_bounds : tuple of arrays, optional
        (f_min, f_max) to normalize F. If None, uses min/max from F.

    Returns
    -------
    float
        Hyper volume value.
    """
    F = np.array(F)

    if normalize:
        if normalization_bounds:
            f_min, f_max = normalization_bounds
        else:
            f_min, f_max = F.min(axis=0), F.max(axis=0)

        F = (F - f_min) / (f_max - f_min + 1e-12)

        if reference_point is None:
            reference_point = np.ones(F.shape[1])
    else:
        if reference_point is None:
            # Conservative fallback: 10% worse than the worst observed
            ref_buffer = 0.1 * np.abs(F.max(axis=0))
            reference_point = F.max(axis=0) + ref_buffer

    hv = HV(ref_point=np.array(reference_point))
    return hv.do(F)




import os
import subprocess
from pathlib import Path
from platform import system
from typing import Union
from apsimNGpy.exceptions import ForgotToRunError, EmptyDateFrameError
import pandas as pd
import matplotlib.pyplot as plt
from sqlalchemy import label
from abc import abstractmethod, ABC
from collections import OrderedDict
from apsimNGpy.settings import logger
from functools import wraps

try:
    import seaborn as sns
except ModuleNotFoundError:
    logger.info("Seaborn is not installed. Please install it to continue.")
    import sys

    sys.exit(1)

HOME = Path.home()


def open_file(filepath):
    """Open a file using the default OS application."""

    if system() == 'Darwin':
        subprocess.call(['open', filepath])
    elif system() == 'Windows':
        os.startfile(filepath)
    elif system() == 'Linux':
        subprocess.call(['xdg-open', filepath])
    else:
        raise OSError('Unsupported operating system')


def inherit_docstring_from(obj):
    def decorator(func):
        @wraps(obj)
        def wrapper(*args, **kwargs):
            return func(*args, **kwargs)

        wrapper.__doc__ = func.__doc__ + "\n" + obj.__doc__
        return wrapper

    return decorator


added_plots = OrderedDict()


class PlotManager(ABC):
    """"
     Abstract base class for high-level visualization of APSIMNGpy simulation outputs.

    The Goal of this class is to provide quick and high level plotting functions for apsimNGpy simulations. Wraps around seaborn and pandas plotting fucntions

    This class provides a clean interface for generating commonly used plots from
    APSIM model results using seaborn and pandas under the hood. It is intended
    to be subclassed by simulation engines (e.g., CoreModel) that implement access
    to APSIM outputs through the abstract methods.

    Features:
        - High-level wrappers for Seaborn plots (lineplot, catplot, scatterplot, histplot, etc.)
        - Automatic result refreshing and cleanup between plots
        - Rendering and saving utility with customizable aesthetics
        - Built-in support for correlation heatmaps and grouped comparisons
        - Leverages APSIM report variables dynamically for plotting

    Requirements for subclasses:
        - Must implement the `results` property
        - Must provide methods: `get_simulated_output`, `add_report_variable`, and `run`

    Example use:
        Subclass PlotManager in your simulation engine and use built-in methods to
        quickly visualize results:
            >>> model = ApsimModel(...)
            >>> model.run()
            >>> model.series_plot(y='Biomass', x='DaysAfterSowing')
            >>> model.render_plot(title='Biomass over Time', show=True, ylabel='Biomass kg/ha', xlabel = 'Days after sowing')

    See Also:
        - seaborn.lineplot
        - seaborn.catplot
        - seaborn.scatterplot
        - pandas.DataFrame.boxplot
    """

    def __init__(self):

        self.added_plots = None
        self.displayed = False

    @property
    @abstractmethod
    def results(self):
        pass

    @abstractmethod
    def get_simulated_output(self, report_name: str):
        pass

    @abstractmethod
    def add_report_variable(self, specification: str):
        pass

    @abstractmethod
    def run(self, report_name: Union[str, list, tuple]):
        pass

    def _clean_numeric_data(self, exclude_vars=('SimulationID', 'CheckpointID', 'CheckpointName')):
        """Select numeric _variables and remove low-signal columns."""
        df = self.results.select_dtypes(include="number").copy()

        # Drop excluded _variables
        for var in exclude_vars:
            if var in df.columns:
                df.drop(columns=var, inplace=True)

        # Remove columns with near-zero sum
        low_info_cols = [col for col in df.columns if df[col].sum() < 1]
        if low_info_cols:
            print(f"Removing low-info columns: {low_info_cols}")
            df.drop(columns=low_info_cols, inplace=True)

        return df

    @inherit_docstring_from(pd.DataFrame)
    def boxplot(self, column, *,
                by=None, figsize=(10, 8), grid=False, **kwargs):

        """
        Plot a boxplot from the simulation results using ``pandas.DataFrame.boxplot`` \n
        =======================================================================.
        """
        self._refresh()

        if self.results is None:
            raise ForgotToRunError("Results not found.")
        df = self.results

        if column not in df.columns:
            raise ValueError(f"Column '{column}' not found in results.")

        if by and by not in df.columns:
            raise ValueError(f"Grouping column '{by}' not found in results.")

        # Plot
        plt.figure(figsize=figsize)
        ax = df.boxplot(column=column, by=by, grid=grid, **kwargs)

        return ax

    @inherit_docstring_from(sns.histplot)
    def distribution(self, x, *, data=None, **kwargs):
        """Plot distribution for a numeric variable. It uses ``seaborn.histplot`` function. Please see their documentation below
        =========================================================================================================\n
        """
        added_plots['current_plot'] = 'distribution'
        self._refresh()

        from pandas.api.types import is_string_dtype

        if is_string_dtype(self.results[x]):
            raise ValueError(f"{x} contains strings")

        df = data or self.results
        sns.histplot(data=df, x=x, kde=True, **kwargs)

    def label(self):
        ...

    def render_plot(
            self,
            *,
            save_as: Union[str, Path] = '',
            dpi: int = 600,
            show: bool = True,
            xlabel: str = '',
            ylabel: str = '',
            title: str = '',
            titlesize: int = 12,
            axessize: int = 14,
            xticksize: int = 10,
            yticksize: int = 10,
            legend_size: int = 12,
            figtitle_size: int = 15
    ):

        # Configure plot aesthetics
        plt.rcParams['axes.titlesize'] = titlesize
        plt.rcParams['axes.labelsize'] = axessize
        plt.rcParams['xtick.labelsize'] = xticksize
        plt.rcParams['ytick.labelsize'] = yticksize
        plt.rcParams['legend.fontsize'] = legend_size
        plt.rcParams['figure.titlesize'] = figtitle_size

        # Set optional labels and title

        if xlabel:
            plt.xlabel(xlabel)
        ylabel
        if ylabel:
            plt.ylabel(ylabel)
        if title:
            plt.title(title)

        plt.tight_layout()

        if save_as:
            plt.savefig(str(save_as), dpi=dpi)

        if show:
            self.show()
            self.displayed = True

    @staticmethod
    def show():
        try:
            if plt.get_fignums():
                plt.show()
            else:
                logger.info('Plot is closed. Initiate again')
        finally:
            if plt.get_fignums():
                plt.close()

    @staticmethod
    def _refresh():
        if plt.get_fignums():
            plt.close()

    @inherit_docstring_from(sns.lineplot)
    def series_plot(self, data=None, *, x: str = None, y: Union[str, list] = None, hue=None, size=None, style=None,
                    units=None, weights=None,
                    palette=None, hue_order=None, hue_norm=None, sizes=None, size_order=None, size_norm=None,
                    dashes=True, markers=None, style_order=None, estimator='mean', errorbar=('ci', 95), n_boot=1000,
                    seed=None, orient='x', sort=True, err_style='band', err_kws=None, legend='auto',
                    ci='deprecated', ax=None, **kwargs):
        """
        Just a wrapper for seaborn.lineplot that supports multiple y columns that could be provided as a list

        Examples::

           from apsimNGpy.core.apsim import ApsimModel
           model = ApsimModel(model= 'Maize')
           # run the results
           model.run(report_names='Report')
           model.series_plot(x='Maize.Grain.Size', y='Yield')
           model.render_plot(show=True, ylabel = 'Maize yield', xlabel ='Maize grain size')

        Plot two variables::

           model.series_plot(x='Yield', y=['Maize.Grain.N', 'Maize.Grain.Size'])

         see below or https://seaborn.pydata.org/generated/seaborn.lineplot.html \n
        =============================================================================================================================================\n
        """
        self._refresh()
        added_plots['current_plot'] = 'series_plot'
        if data is None:
            data = self.results

        df = data.copy()

        if isinstance(y, list):
            if x is None:
                raise ValueError("When passing a list for y, x must be specified")

            # Ensure x column exists
            if x not in df.columns:
                raise KeyError(f"{x} not found in data")

            # Melt into long-form
            df = df[[x] + y].melt(id_vars=x, var_name='Variable', value_name='Value')

            # Adjust parameters for seaborn
            x = x
            y = 'Value'
            if hue is None:
                hue = 'Variable'  # Color lines by variable name

        sns.lineplot(
            data=df, x=x, y=y, hue=hue, size=size, style=style, units=units, weights=weights,
            palette=palette, hue_order=hue_order, hue_norm=hue_norm, sizes=sizes, size_order=size_order,
            size_norm=size_norm,
            dashes=dashes, markers=markers, style_order=style_order, estimator=estimator, errorbar=errorbar,
            n_boot=n_boot,
            seed=seed, orient=orient, sort=sort, err_style=err_style, err_kws=err_kws, legend=legend,
            ci=ci, ax=ax, **kwargs)

    @inherit_docstring_from(sns.scatterplot)
    def scatter_plot(
            self,
            data=None,
            *,
            x=None,
            y=None,
            hue=None,
            size=None,
            style=None,
            palette=None,
            hue_order=None,
            hue_norm=None,
            sizes=None,
            size_order=None,
            size_norm=None,
            markers=True,
            style_order=None,
            legend='auto',
            ax=None,
            **kwargs
    ):
        """
        Plot scatter plot using seaborn with flexible aesthetic mappings.
        reference: https://seaborn.pydata.org/generated/seaborn.scatterplot.html. Check seaborn documentation below for more details \n
        ================================================================================================================================\n"""
        self._refresh()
        if data is None:
            data = self.results
        sns.scatterplot(
            data=data,
            x=x,
            y=y,
            hue=hue,
            size=size,
            style=style,
            palette=palette,
            hue_order=hue_order,
            hue_norm=hue_norm,
            sizes=sizes,
            size_order=size_order,
            size_norm=size_norm,
            markers=markers,
            style_order=style_order,
            legend=legend,
            ax=ax,
            **kwargs
        )

    @inherit_docstring_from(sns.catplot)
    def cat_plot(self,
                 data=None,
                 *,
                 x=None,
                 y=None,
                 hue=None,
                 row=None,
                 col=None,
                 kind='strip',
                 estimator='mean',
                 errorbar=('ci', 95),
                 n_boot=1000,
                 seed=None,
                 units=None,
                 weights=None,
                 order=None,
                 hue_order=None,
                 row_order=None,
                 col_order=None,
                 col_wrap=None,
                 height=5,
                 aspect=1,
                 log_scale=None,
                 native_scale=False,
                 formatter=None,
                 orient=None,
                 color=None,
                 palette=None,
                 hue_norm=None,
                 legend='auto',
                 legend_out=True,
                 sharex=True,
                 sharey=True,
                 margin_titles=False,
                 facet_kws=None,
                 **kwargs
                 ):
        """Wrapper for seaborn.catplot with all keyword arguments.
        reference https://seaborn.pydata.org/generated/seaborn.catplot.html or check seaborn documentation below\n
        =========================================================================================================\n"""
        self._refresh()
        added_plots['cat_plot'] = 'cat_plot'
        df = self.results if data is None else data
        return sns.catplot(
            data=df,
            x=x,
            y=y,
            hue=hue,
            row=row,
            col=col,
            kind=kind,
            estimator=estimator,
            errorbar=errorbar,
            n_boot=n_boot,
            seed=seed,
            units=units,
            weights=weights,
            order=order,
            hue_order=hue_order,
            row_order=row_order,
            col_order=col_order,
            col_wrap=col_wrap,
            height=height,
            aspect=aspect,
            log_scale=log_scale,
            native_scale=native_scale,
            formatter=formatter,
            color=color,
            palette=palette,
            hue_norm=hue_norm,
            legend=legend,
            legend_out=legend_out,
            sharex=sharex,
            sharey=sharey,
            margin_titles=margin_titles,
            facet_kws=facet_kws,
            **kwargs
        )

    def relplot(self, data=None, **kwargs):
        data = data or self.results
        g = sns.relplot(data=data, **kwargs)
        return g

    def correlation_heatmap(self, columns: list = None, figsize=(10, 8), **kwargs):
        self._refresh()
        added_plots['correlation_heatmap'] = 'correlation_heatmap'
        """Plot correlation heatmap for numeric _variables."""
        if columns:
            df = self.results[columns]
        else:
            df = self._clean_numeric_data()
            if df.empty:
                raise EmptyDateFrameError("No valid numeric data for correlation heatmap.")

        corr = df.corr()
        plt.figure(figsize=figsize)
        sns.heatmap(corr, annot=True, cmap="coolwarm", fmt=".2f", **kwargs)
        plt.title("Correlation Heatmap")
        plt.tight_layout()


if __name__ == '__main__':
    from apsimNGpy.core.apsim import ApsimModel
    from apsimNGpy.core.base_data import load_default_simulations

    scrach = Path.home() / 'testss'
    scrach.mkdir(exist_ok=True)

    apsim = load_default_simulations(crop='Maize', simulations_object=False, set_wd=scrach)
    model = ApsimModel(model='maize')
    model.add_report_variable(variable_spec='[Soil].Nutrient.TotalC[1]/1000 as SOC1', report_name='Report',
                              set_event_names=['[Clock].EndOfYear', '[Maize].Harvesting'])
    model.add_db_table(
        variable_spec=['[Soil].Nutrient.TotalC[1]/1000 as SOC1', '[Soil].Nutrient.TotalC[2]/1000 as SOC2',
                       '[Clock].Today.Year as Year'])
    # model.add_db_table(variable_spec=['[Soil].Nutrient.TotalC[2]/1000 as SOC2',])
    model.update_mgt(management=({"Name": 'Sow using a variable rule', 'Population': 8},))
    model.run(report_name='my_table')

    model.cat_plot(y='SOC1', x='Zone')

    model.series_plot(y='SOC2', x='Year')
    model.distribution('SOC1', show=True)
    # model.plot_distribution('Yield')
    model.run(report_name='Report')
    model.correlation_heatmap()
    # model.get_weather_from_web(lonlat=(-93.50456, 42.601247), start=1990, end=2001, source='daymet')



from typing import Union
from dataclasses import dataclass
import numpy as np

from pymoo.algorithms.moo.nsga2 import NSGA2
from pymoo.config import Config
from pymoo.core.problem import ElementwiseProblem
from pymoo.optimize import minimize
from pymoo.core.problem import Problem
from pymoo.algorithms.moo.nsga2 import NSGA2
from pymoo.algorithms.moo.nsga3 import NSGA3
from pymoo.optimize import minimize
from apsimNGpy.core.cal import OptimizationBase as Runner, OptimizationBase
import numpy as np
from apsimNGpy.optimizer.optutils import compute_hyper_volume, edit_runner
from apsimNGpy.optimizer.base import AbstractProblem
from pymoo.core.problem import ElementwiseProblem
from apsimNGpy.core.cal import OptimizationBase as Runner
import numpy as np

# Disable compilation warning
Config.warnings['not_compiled'] = False

class MultiObjectiveProblem(AbstractProblem):

    def __init__(self, apsim_model: Runner, objectives: list, *, decision_vars: list = None, cache_size = 100):
        """
                   Parameters
                   ----------
                   apsim_runner : apsimNGpy.core.cal.OptimizationBase
                       Instance to run APSIM simulations.
                   objectives : list of callable
                       List of functions that take simulation output (DataFrame) and return scalar objective values.
                   decision_vars : list of dict, optional
                       Each dict must have: 'path', 'bounds', 'v_type', 'kwargs'.
                   """
        AbstractProblem.__init__(self, apsim_model, max_cache_size=cache_size)
        self.apsim_model = apsim_model
        self.objectives = objectives
        self.params = []
        self.decision_vars = decision_vars if decision_vars is not None else []
        self.optimization_type = 'multi-objective'

        assert all(callable(obj) for obj in objectives), "All objectives must be callable"


    def evaluate_objectives(self, x, *args, **kwargs):
        for val, spec in zip(x, self.decision_vars):
            edit_runner(self.apsim_model, decision_specs=spec, x_values=val)
        # get results
        df = self.apsim_model.run().results
        return np.array([obj(df) for obj in self.objectives])

    def  optimization_type(self):
         return 'multi-objective'

    def minimize_with_local_solver(self, **kwargs):
        pass  # not needed in this problem

    class SetUpProblem(ElementwiseProblem):
        def __init__(self, core_problem, **kwargs):
            self.core_problem = core_problem
            xl, xu = core_problem.extract_bounds()
            super().__init__(
                n_var=len(core_problem.decision_vars),
                n_obj=len(core_problem.objectives),
                n_ieq_constr=0,
                xl=np.array(xl),
                xu=np.array(xu), *kwargs
            )

        def _evaluate(self, x, out, *args, **kwargs):
            out["F"] = self.core_problem.evaluate_objectives(x, args, kwargs)

    def extract_bounds(self):
        xl, xu = [], []
        for var in self.decision_vars:
            if var['v_type'] not in {'int', 'float'}:
                raise ValueError(f"Unsupported variable type: {var['v_type']}")
            xl.append(var['bounds'][0])
            xu.append(var['bounds'][1])
        return xl, xu

    def get_problem(self):
        return self.SetUpProblem(self)

    def is_mixed_type_vars(self):
        """Detect if decision vars contain types other than float or int."""
        return any(var['v_type'] not in {'int', 'float'} for var in self.decision_vars)

    def minimize(self, **kwargs):
        return minimize

    def NSG2(self):
        return NSGA2


# 1. Setup APSIM model
runner = Runner("Maize")
runner.add_report_variable('[Soil].Nutrient.NO3.kgha[1] as nitrate', report_name='Report')

# 2. Define decision variables
decision_vars = [
    {'name': '.Simulations.Simulation.Field.Fertilise at sowing.Amount', 'v_type': 'float', 'bounds': [50, 300]},
    {'name': '.Simulations.Simulation.Field.Sow using a variable rule.Population', 'v_type': 'float', 'bounds': [4, 14]}
]

# below enables defining other parameters that will be required for editing
_vars = [
    {'path': '.Simulations.Simulation.Field.Fertilise at sowing', 'Amount': "?", "bounds": [50, 300],
     "v_type": "float"},
    {'path': '.Simulations.Simulation.Field.Sow using a variable rule', 'Population': "?", 'v_type': 'float',
     'bounds': [4, 14]}
]


# 3. Define objective functions
def negative_yield(df):
    return -df['Yield'].mean()


def nitrate_leaching(df):
    return df['nitrate'].sum()


objectives = [negative_yield, nitrate_leaching]























import pandas as pd
from apsimNGpy.core.pythonet_config import is_file_format_modified
from dataclasses import dataclass
from gc import collect
import Models
from System.Collections import IEnumerable
from System import String, Double, Array
from functools import lru_cache, cache
from typing import Union, Dict, Any
from Models.Soils import Physical, SoilCrop, Organic, Solute, Chemical
import warnings
from System.Collections.Generic import List, KeyValuePair
from apsimNGpy.settings import *
from pathlib import Path
from apsimNGpy.core.model_loader import load_apsim_model
from apsimNGpy.core.cs_resources import simple_rotation_code, update_manager_code
from apsimNGpy.core.config import apsim_version

IS_NEW_APSIM = is_file_format_modified()

from apsimNGpy.core.cs_resources import CastHelper, sow_using_variable_rule, sow_on_fixed_date, harvest, \
    fertilizer_at_sow, cast_as

APSIM_VERSION = apsim_version(True)
BASE_RELEASE_NO = '2025.8.7837.0'


@cache
def select_thread(multithread):
    if multithread:
        runtype = Models.Core.Run.Runner.RunTypeEnum.MultiThreaded
    else:
        runtype = Models.Core.Run.Runner.RunTypeEnum.SingleThreaded
    return runtype


select_thread(multithread=True)


def _tools(method):
    if hasattr(Models.Core.ApsimFile.Structure, 'Rename'):
        rename = Models.Core.ApsimFile.Structure.Rename
    else:
        rename = None
    config = {
        "ADD": Models.Core.ApsimFile.Structure.Add,
        "DELETE": Models.Core.ApsimFile.Structure.Delete,
        "MOVE": Models.Core.ApsimFile.Structure.Move,
        "RENAME": rename,
        "CLONER": Models.Core.Apsim.Clone,
        "REPLACE": Models.Core.ApsimFile.Structure.Replace,
        "MultiThreaded": Models.Core.Run.Runner.RunTypeEnum.MultiThreaded,
        "SingleThreaded": Models.Core.Run.Runner.RunTypeEnum.SingleThreaded,
        "ModelRUNNER": Models.Core.Run.Runner,
        "CLASS_MODEL": type(Models.Clock),
        "ACTIONS": ('get', 'delete', 'check'),
        "COLLECT": collect,
    }
    method = config[method]
    del config
    return method


CLASS_MODEL = _tools('CLASS_MODEL')


def find_child(parent, child_class, child_name):
    parent = getattr(parent, 'Model', parent)
    if isinstance(child_class, str):
        child_class = validate_model_obj(child_class)

    for child in parent.Children:
        cast_child = CastHelper.CastAs[child_class](child)
        if child.Name == child_name and cast_child:
            # child = CastHelper.CastAs[child_class](child)
            return child
        # 🔁 Recursively search in child's children
        result = find_child(child, child_class, child_name)
        if result is not None:
            return result

    return None  # Not found


def find_all_in_scope(parent, child_class):
    """

    @param parent: base model scope to search
    @param child_class: Model class from the Models namespace to search for
    @return: list[Models.Core.IModel] objects
    """

    all_in_scope = []
    if isinstance(child_class, str):
        child_class = validate_model_obj(child_class)

    for child in parent.Children:
        cast_child = CastHelper.CastAs[child_class](child)

        if cast_child:
            all_in_scope.append(cast_child)

        # 🔁 Recursively search in this child's children
        child_results = find_all_in_scope(child, child_class)
        if child_results:
            all_in_scope.extend(child_results)
    return all_in_scope


def find_descendant(parent, child_class):

    parent = getattr(parent, 'Model', parent)
    if isinstance(child_class, str):
        child_class = validate_model_obj(child_class)
    for child in parent.Children:
        cast_child = CastHelper.CastAs[child_class](child)
        if cast_child:
            # child = CastHelper.CastAs[child_class](child)
            return child
        # 🔁 Recursively search in child's children
        result = find_descendant(child, child_class)
        if result is not None:
            return result


def find_child_of_class(parent, child_class):
    """
    Finds a child with a break. Equivalent to old method of FindDescendant. can return None if no child

    """
    return find_descendant(parent, child_class)


def get_or_check_model(search_scope, model_type, model_name, action='get', cache_size=300):
    """
            Helper function to check if a model instance is found in the simulation
            and perform a specified action.

            Actions supported:
            - 'get': returns the found model.
            - 'delete': deletes the model.
            - 'check': returns True/False depending on whether the model exists.

            Parameters:
            -----------
            sim : ApsimSimulation
                The APSIM simulation object.
            model_class : str
                The type of model to search for (e.g., 'Soil', 'Manager').
            model_name : str, optional
                The name of the model to find. If omitted, returns the first model of that type.
            action : str
                One of 'get', 'delete', or 'check'.

            Returns:
            --------
            object or bool or None
                - The found model if action is 'get'.
                - True/False if action is 'check'.
                - None if action is 'delete'.

            Raises:
            -------
            ValueError
                If the model is not found (when action is 'get' or 'delete'),
                or if an invalid action is specified.
            """

    def _execute(search_scope, model_type, model_name, action):
        if action not in ModelTools.ACTIONS:
            raise ValueError(f'sorry action should be any of {ModelTools.ACTIONS} ')
        # get bound methods based on model type
        try:
            finder = search_scope.FindIDescendant[model_type]
            get_model = finder(model_name) if model_name else finder()
        except AttributeError:
            get_model = find_child(search_scope, child_class=model_type, child_name=model_name)
        if action == 'check':
            return True if get_model is not None else False
        if not get_model and action == 'get':
            if model_type == 'Models.Core.Simulations' or model_type == Models.Core.Simulations:
                return search_scope
            raise ValueError(f"{model_name} of type {model_type} not found")

        if action == 'delete' and get_model:
            ModelTools.DELETE(get_model)
        if get_model and action == 'get':
            return get_model

    # if cacheit:
    #     __excute = lru_cache(maxsize=cache_size)(__excute)
    return _execute(search_scope, model_type, model_name, action)


def _find_model(model_name: str, model_namespace=Models, target_type=CLASS_MODEL) -> CLASS_MODEL:
    """
    Recursively find a model by name within a namespace.

    Args:
        model_name (str): The name of the model to find.
        model_namespace (object, optional): The root namespace (defaults to Models).
        target_type (type, optional): Expected type of the model.

    Returns:
        object: The found model object, or None if not found.
    """

    if not hasattr(model_namespace, "__dict__"):
        return None

    ITEMS = model_namespace.__dict__
    if isinstance(model_name, str):
        model_name = model_name.capitalize()
    for attr, value in ITEMS.items():
        if attr == model_name and isinstance(value, target_type):
            return value

        _get_attr = getattr(value, model_name, None)

        if isinstance(_get_attr, target_type):
            return _get_attr

        if hasattr(value, "__dict__"):

            result = _find_model(model_name, value, target_type=target_type)
            if result:
                return result
    # recursion is complete but model is none

    return None


def find_model(model_name: str):
    model_type = _find_model(model_name)
    if model_type:
        return model_type
    # for some unknown reasons Models.Factorial namespace is not featuring so let's do it mannually here
    import Models
    if model_type is None:
        model_type = getattr(Models.Factorial, model_name, None)
    if model_type is None:
        model_type = getattr(Models.Surface, model_name, None)
    if model_type is None:
        model_type = getattr(Models.Core, model_name, None)
    if model_type is None:
        model_type = getattr(Models.Storage, model_name, None)
    collect()
    return model_type


@lru_cache(maxsize=300)
def validate_model_obj(model__type, evaluate_bound=False) -> CLASS_MODEL:
    """
    Evaluates the model type from either string or Models namespace
    @param model__type (str,Models, required)
    it can also accept either string or Models namespace as single letter specifying the final model or as a whole path e.g 'Models.Clock' and 'Clock' are all valid
         'Clock', 'Simulation', 'Manager', 'Report','Simulations', 'Weather', 'Soil
    @return:
       model type from Models namespace

    :Raises
       raises a ValueError if model__type is not valid or found from the Models namespace
    """
    model_types = None
    bound_model = None

    try:
        if isinstance(model__type, str):
            _model_name_space = 'Models.'
            ln = len(_model_name_space)
            # find _find_model could be sufficent, where we only allow the user to supply a single name but we dont know whether some names repeated on some models
            if find_model(model__type):
                from_single = find_model(model__type)
                if from_single is not None:
                    return from_single
            if model__type[:ln] == _model_name_space:
                _model_type = eval(model__type)
                bound_model = _model_type.__class__
                if isinstance(_model_type, ModelTools.CLASS_MODEL):
                    model_types = _model_type

        if isinstance(model__type, ModelTools.CLASS_MODEL):
            model_types = model__type

        if evaluate_bound and not isinstance(model__type, str):
            bound_model = model__type.__class__
        if isinstance(bound_model, ModelTools.CLASS_MODEL):
            model_types = bound_model

        if model_types:
            return model_types
        else:
            raise ValueError(f"invalid model_class: '{model__type}' from type: {type(model__type)}")
    finally:
        # future implementation

        pass


def extract_value(model_instance, parameters=None):
    if isinstance(parameters, str):
        parameters = {parameters}

    match type(model_instance):
        case Models.Climate.Weather:
            value = model_instance.FileName
        case Models.Clock:
            import datetime
            def __convert_to_datetime(dotnet_dt):
                return datetime.datetime(
                    dotnet_dt.Year,
                    dotnet_dt.Month,
                    dotnet_dt.Day,
                    dotnet_dt.Hour,
                    dotnet_dt.Minute,
                    dotnet_dt.Second,
                    dotnet_dt.Millisecond * 1000  # Python uses microseconds
                )

            validated = dict(End='End', Start='Start', end='End', start='Start', end_date='End', start_date='Start')
            accepted_attributes = {'Start', 'End'}
            selected_parameters = {validated.get(k) for k in parameters if
                                   validated.get(k) in accepted_attributes} if parameters else set()
            dif = accepted_attributes - selected_parameters
            if dif == accepted_attributes and parameters:
                raise ValueError(
                    f"To inspect the 'Clock Model Parameters:\n, Parameters must be None or any of '{', '.join(validated.keys())}'")
            attributes = selected_parameters or accepted_attributes

            if len(attributes) == 1:
                value = __convert_to_datetime(getattr(model_instance, *attributes))
            else:
                value = {atr: __convert_to_datetime(getattr(model_instance, atr)) for atr in attributes}

        case Models.Manager:
            selected_parameters = parameters if parameters else set()

            if selected_parameters:
                value = {param.Key: param.Value for param in model_instance.Parameters if
                         param.Key in selected_parameters}
            else:
                value = {param.Key: param.Value for param in model_instance.Parameters}
        case Models.Soils.Physical | Models.Soils.Chemical | Models.Soils.Organic | Models.Soils.Water | Models.Soils.Solute:
            # get selected parameters
            selected_parameters = {k for k in parameters if hasattr(model_instance, k)} if parameters else set()

            thick = getattr(model_instance, 'Thickness', None)

            df = pd.DataFrame()  # empty data frame
            attributes = selected_parameters or dir(model_instance)
            evp = [at for at in attributes if not at.startswith('__')]
            if not selected_parameters and parameters:
                raise ValueError(f"Parameters must be none or any of '{", \n".join(evp)}'")
            if thick:
                for attr in attributes:
                    if attr.startswith('__'):
                        continue
                    val = getattr(model_instance, attr)
                    if isinstance(val, IEnumerable) and not isinstance(val, String):
                        val_list = list(val)
                        if len(val_list) == len(thick):
                            df[attr] = val_list
            value = df
        case Models.Report:
            accepted_attributes = {'VariableNames', 'EventNames'}
            selected_parameters = {k for k in parameters if hasattr(model_instance, k)} if parameters else set()
            dif = accepted_attributes - selected_parameters
            if dif == accepted_attributes and parameters:
                raise ValueError(f"Parameters must be none or any of '{accepted_attributes}'")

            attributes = selected_parameters or accepted_attributes
            value = {}
            for attr in attributes:
                value[attr] = list(getattr(model_instance, attr))

        case Models.PMF.Cultivar:
            selected_parameters = set(parameters) if parameters else set()
            params = {}
            for cmd in model_instance.Command:
                if cmd:
                    if '=' in cmd:
                        key, val = cmd.split('=', 1)
                        params[key.strip()] = val.strip()
            if selected_parameters:
                value = {k: v for k, v in params.items() if k in selected_parameters}
                if not value:
                    raise ValueError(
                        f"None of '{selected_parameters}' was found. Available parameters are: '{', '.join(params.keys())}'")
            else:
                value = params
        case Models.Surface.SurfaceOrganicMatter:
            selected_parameters = set(parameters) if parameters else set()
            accepted_attributes = {'LyingWt', 'N', 'NH4', 'NO3', 'LabileP', 'SurfOM', 'P', "C", 'Cover',
                                   'InitialCPR', 'InitialResidueMass', 'StandingWt',
                                   'InitialCNR', 'InitialCPR'}
            dif = accepted_attributes - selected_parameters
            if len(dif) == len(accepted_attributes) and parameters:
                raise ValueError(f"Parameters must be none or any of '{', '.join(accepted_attributes)}'")
            params = {}
            attributes = selected_parameters or accepted_attributes
            for attrib in attributes:
                x_attrib = getattr(model_instance, attrib)
                params[attrib] = x_attrib

            value = params
        case _:
            raise NotImplementedError(f"No inspect input method implemented for model type: {type(model_instance)}")
    return value


def inspect_model_inputs(scope, model_type: str, model_name: str,
                         simulations: Union[str, list] = MissingOption,
                         parameters: Union[list, str] = None,
                         **kwargs) -> Union[Dict[str, Any], pd.DataFrame, list, Any]:
    """
    Inspect the current input values of a specific APSIM model type instance within given simulations.

    Parameters:
    -----------
    scope : ApsimNG model context with Models.Core.Simulations  and its associated Simulation
        Root scope or project object containing Simulations.
    model_class : str
        Name of the model class (e.g., 'Clock', 'Manager', 'Soils.Physical', etc.)
    simulations : Union[str, list]
        Name or list of names of simulation(s) to inspect.
    model_name : str
        Name of the model instance within each simulation.
    **kwargs : dict
        Optional keyword arguments — not used here but accepted for interface compatibility.

    Returns:
    --------
    Union[Dict[str, Any], pd.DataFrame, list, Any]
        - For Weather: file path(s)
        - For Clock: (start, end) dict(s) or datetime if any of parameters is one. To narrow down the inspection, accepted parameters are, Start and End supplied as a list or a str if any of them is needed,
        - For Manager: dictionary of parameters
        - For Soil models: pandas DataFrame(s) of layer-based properties
        - For Report: dictionary with 'VariableNames' and 'EventNames'. To narrow down the inspection, use 'VariableNames' and 'EventNames' supplied as a list or a str if any of them is needed,
        - For Cultivar: dictionary of parsed parameter=value pairs

    Raises:
    -------
    ValueError:
        If model is not found or invalid arguments are passed.
    NotImplementedError:
        If the model type is unsupported.

    Requirements:
    -------------
    - APSIM Next Gen Python bindings (apsimNGpy)
    - Python 3.10+
    """

    model_type_class = validate_model_obj(model_type)

    is_single_sim = True if isinstance(simulations, str) else False
    sim_list = scope.find_simulations(simulations)

    result = {} if not is_single_sim else None
    if isinstance(parameters, str):
        parameters = {parameters}
    for sim in sim_list:
        model_instance = find_child(sim, child_class=model_type_class, child_name=model_name)
        model_class = validate_model_obj(model_type_class)
        model_instance = CastHelper.CastAs[model_class](model_instance)
        value = extract_value(model_instance, parameters)
        if is_single_sim:
            return value
        else:
            result[sim.Name] = value

    return result


def replace_variable_by_index(old_list: list, new_value: list, indices: list):
    for idx, new_val in zip(indices, new_value):
        old_list[idx] = new_val
    return old_list


@lru_cache(maxsize=None)
def soil_components(component):
    _comp = component.lower()
    comps = {'organic': Organic,
             'physical': Physical,
             'soilcrop': SoilCrop,
             'solute': Solute,
             'chemical': Chemical,
             }
    return comps[_comp]


def old_method(old_method, new_method):
    warnings.warn(
        f"{old_method} is deprecated and will be removed in a future release/version. Please use {new_method}() instead.",
        category=DeprecationWarning,
        stacklevel=2
    )
    # method logic here


def run_p(_model, simulations='all', clean=False, multithread=True):
    """Run apsim model in the simulations

    Parameters
    ----------
    ``_model`` : ApsimNGpy model object

    ``simulations`` (__str_), optional
        List of simulation names to run, if `None` runs all simulations, by default `None`.

    ``clean`` (_-boolean_), optional
        If `True` remove existing database for the file before running, deafults to False`.

    ``multithread``, optional
        If `True` APSIM uses multiple threads, by default `True`.

    """

    if clean:
        _model._DataStore.Dispose()
        Path(_model._DataStore.FileName).unlink(missing_ok=True)
        _model._DataStore.Open()

    if simulations == 'all':
        simulations = None
    sims = _model.find_simulations(simulations)  # returns a list
    print(sims)
    # Runner needs C# list
    cs_sims = List[Models.Core.Simulation]()
    for sim in sims:
        cs_sims.Add(sim)
    runmodel = Models.Core.Run.Runner(cs_sims, True, False, False, None, select_thread(multithread))
    e = runmodel.Run()
    if (len(e) > 0):
        logger.info(e[0].ToString())


def _edit_in_cultivar(_model, model_name, param_values, simulations=None, verbose=False):
    # if there is a need to target a specific simulation in case of changing the current culvar in simulation

    _cultivar_names = _model.inspect_model('Cultivar', fullpath=False)

    # Extract input parameters
    commands = param_values.get("commands")
    values = param_values.get("values")

    cultivar_manager = param_values.get("cultivar_manager")
    new_cultivar_name = param_values.get("new_cultivar_name", None)
    cultivar_manager_param = param_values.get("parameter_name", 'CultivarName')
    plant_name = param_values.get('plant')
    # Input validation
    required_keys = ["commands", "values", "cultivar_manager", "parameter_name", "new_cultivar_name"]
    # Extract input parameters
    missing = [key for key in required_keys if not param_values.get(key)]

    if missing:
        raise ValueError(f"Missing required parameter(s): {', '.join(missing)}")

    if isinstance(commands, (list, tuple)) or isinstance(values, (list, tuple)):
        assert isinstance(commands, (list, tuple)) and isinstance(values, (list, tuple)), \
            ("Both `commands` and `values` must be iterables (list or tuple), not sets or mismatched "
             "types")
        assert len(commands) == len(values), "`commands` and `values` must have the same length"

    # Get replacement folder and source cultivar model
    replacements = get_or_check_model(_model.Simulations, Models.Core.Folder, 'Replacements',
                                      action='get')

    get_or_check_model(replacements, Models.Core.Folder, 'Replacements',
                       action='delete')
    cultivar_fallback = get_or_check_model(_model.Simulations, Models.PMF.Cultivar, model_name,
                                           action='get', )
    cultivar = ModelTools.CLONER(cultivar_fallback)

    # Update cultivar parameters
    cultivar_params = _model._cultivar_params(cultivar)

    if isinstance(values, str):
        cultivar_params[commands] = values.strip()
    elif isinstance(values, (int, float)):
        cultivar_params[commands] = values
    else:
        for cmd, val in zip(commands, values):
            cultivar_params[cmd.strip()] = val.strip() if isinstance(val, str) else val

    # Apply updated commands
    updated_cmds = [f"{k.strip()}={v}" for k, v in cultivar_params.items()]
    cultivar.set_Command(updated_cmds)

    # Attach cultivar under a plant model
    plant_model = get_or_check_model(replacements, Models.PMF.Plant, plant_name,
                                     action='get')

    # Remove existing cultivar with same name
    get_or_check_model(replacements, Models.PMF.Cultivar, new_cultivar_name, action='delete')

    # Rename and reattach cultivar
    cultivar.Name = new_cultivar_name
    ModelTools.ADD(cultivar, plant_model)

    for sim in _model.find_simulations(simulations):
        # Update cultivar manager script
        _model.edit_model(model_type=Models.Manager, simulations=sim.Name,
                          model_name=cultivar_manager, **{cultivar_manager_param: new_cultivar_name})

    if verbose:
        logger.info(f"\nEdited Cultivar '{model_name}' and saved it as '{new_cultivar_name}'")
    _model.save()


def load_parameters(**kwargs):
    params = List[KeyValuePair[String, String]]()
    for k, v in kwargs.items():
        kv = KeyValuePair[String, String](String(k), String(str(v)))
        params.Add(kv)
    return params


def add_model(_model, parent, model):
    parent = validate_model_obj(parent)
    if callable(model):
        model = model()
    find_parent = _model.Simulations


def add_as_simulation(_model, resource, sim_name):
    """
    Get a simulation from another file or model and it to the current apsimNGpy model simulation
    _model: apsimNgpy.core.apsim.ApsimModel,
    resource: source of the simulation file,
    sim_name, new name to name the added simulation
    """
    model = load_apsim_model(resource)
    if APSIM_VERSION > BASE_RELEASE_NO:
        sim = find_child_of_class(model.IModel, Models.Core.Simulation)
    else:
        sim = model.IModel.FindDescendant[Models.Core.Simulation]()
    if sim is None:
        raise RuntimeError(f'simulation not found {model.IModel}')
    new_sim = sim
    names = _model.inspect_model(Models.Core.Simulation, fullpath=False)
    if sim_name in names:
        raise ValueError(f"new '{sim_name}' can not be any of '{",".join(names)}'")
    try:
        if IS_NEW_APSIM:
            try:
               new_sim.Rename(sim_name)
            except AttributeError:
                new_sim.Name = sim_name

        else:
            new_sim.Name = sim_name
        model.Simulations.sim_name = sim_name
        _model.Simulations.Children.Add(new_sim)

        ModelTools.DELETE(sim)
        return model
    finally:
        os.remove(model.path)


def detect_sowing_managers(_model):
    if APSIM_VERSION > BASE_RELEASE_NO:
        managers = find_all_in_scope(_model.Simulations, Models.Manager)
    else:
        managers = _model.Simulations.FindAllDescendants[Models.Manager]()
    if managers is not None:
        for manager in managers:
            code = manager.Code
            if code is not None:
                if 'Crop.Sow' in code and 'using Models.PMF' in code and 'IPlant' in code:
                    return manager


def compile_manager(code):
    manager = Models.Manager()
    manager.set_Code(code)
    return manager


planting_info = {'seq': 'Maize, Soybean', 'in_crop': ('Maize',)}


def configure(app, seq: Union[str, tuple, list], in_crop: Union[tuple, list], strategy: str,
              sowing_depth: str, row_spacing: str, population: int,
              cultivar_name: str, start_date: str, end_date=None, min_esw=100,
              min_rain=10, rain_duration=4, simulations=None, fertilize_crop=None):
    """

    @param fertilize_crop:
    @param simulations: simulation to configure rotation
    @param min_esw:
    @param min_rain:
    @param sowing_depth:
    @param row_spacing: spacing between rows
    @param population: int, planting density
    @param rain_duration:
    @param cultivar_name: name of the cultivar
    @param app: apsimNGpy model
    @param seq: str e.g (Maize, Soybean) or as a string 'Maize, Soybean'. The first crop will be planted first in the first yea, then followed by another crop after the comma
    @param in_crop: is the crop currently in the simulation
    @param strategy: either fixed or variable, variable required start and end date, while foxed only start date is required
    @param start_date: sowing date or start date if sowing is through a variable time window
    @param end_date: date to end the sowing
    @return:
    """

    def _add_child_model(parent, child_model, name):
        """
        if child is found remove and add, or add, proceed with caution
        @param parent: parent model
        @param child_model:  child imodel
        @param name: name of the child
        @return:
        """

        # enforce
        child_model.Name = name
        ch = parent.Children
        for child in ch:
            if child.Name == name:
                parent.Children.Remove(child)
                parent.Children.Add(child_model)
                break
        else:
            parent.Children.Add(child_model)

    if not fertilize_crop:
        fertilize_crop = {}
    fixed_parameters = dict(SowDate=start_date,
                            SowingDepth=sowing_depth,
                            CultivarName=cultivar_name,
                            RowSpacing=row_spacing,
                            Population=population,
                            Crop="")
    if strategy == 'fixed':
        sow_params = fixed_parameters
    else:
        if end_date is None:
            raise ValueError(f"end_date must be specified if sowing strategy is variable not  fixed")
        fixed_parameters.pop('SowDate', None)
        sow_params = dict(**fixed_parameters, EndDate=end_date, StartDate=start_date, MinESW=min_esw,
                          MinRain=min_rain, RainDays=rain_duration, )
    rotation = '.'.join(seq) if not isinstance(seq, str) else seq
    if isinstance(seq, str):
        seq = seq.split(",")
    in_crop = [in_crop] if isinstance(in_crop, str) else in_crop
    for crop in in_crop:
        if crop not in seq:
            raise ValueError(f"in_crop must be in {seq}")
    sims = app.find_simulations(None)
    # get crop that needs new sowing_manager
    need_sower = set(seq) - set(seq).difference(in_crop)
    print(need_sower)
    print(need_sower)
    # add rotation manager
    rot_manager = compile_manager(simple_rotation_code)
    # set the rotation
    rot_manager.set_Parameters(load_parameters(Crops=rotation))

    for crops in need_sower:
        # add new plant
        m_crop = Models.PMF.Plant()
        m_crop.Name = crops
        m_crop.ResourceName = crops

        for sim in sims:
            zone = sim.FindDescendant[Models.Core.Zone]()
            _add_child_model(zone, rot_manager, 'Simple Rotation Manager')
            _add_child_model(zone, m_crop, crops)
            # add sowing manager
            code = sow_on_fixed_date if strategy == 'fixed' else sow_using_variable_rule
            sow_manager = compile_manager(code)
            sow_manager.Name = f"{crops}_sow_on_fixed_date" if strategy == 'fixed' else f"{crops}_sow_using_variable_rule"
            sow_params['Crop'] = crops
            sow_params['RotationManager'] = rot_manager.Name
            sow_manager.set_Parameters(load_parameters(**sow_params))
            zone.Children.Add(sow_manager)

            # add harvest manager
            manager_harvest = compile_manager(harvest)
            manager_harvest.set_Parameters(load_parameters(Crop=crops))
            manager_harvest.Name = f"harvest{crops}"
            _add_child_model(zone, manager_harvest, manager_harvest.Name)

            # add soil crop
            soil_crop = zone.FindDescendant[Models.Soils.SoilCrop]()
            clonedSoilCrop = ModelTools.CLONER(soil_crop)
            # change its name
            clonedSoilCrop.Name = f"{crops}Soil"
            soilPhysical = zone.FindDescendant[Models.Soils.Physical]()
            for child in soilPhysical.Children:
                if child.Name == f"{crops}Soil":
                    soilPhysical.Children.Remove(child)
            soilPhysical.Children.Add(clonedSoilCrop)
            n_amount = fertilize_crop.get(crops)
            if n_amount:
                # add fertilizer manager or operations
                f_manager = compile_manager(fertilizer_at_sow)
                params = {"Crop": crops, 'Amount': n_amount, 'FertiliseDepth': '10', 'FertiliserType': "UreaN"}
                f_manager.set_Parameters(load_parameters(**params))
                f_manager.Name = f"{crops}-fertilize-at-sowing"
                zone.Children.Add(f_manager)

                app.save()
                # app.edit_model(model_type='Models.Manager',
                #                model_name=f_manager.Name, simulations=sim.Name, **params)

                return find_child(app.Simulations, child_class=Models.Manager, child_name=f_manager.Name)
            ...


def configure_rotation(_model, simulations=None):
    manager = Models.Manager()
    manager.set_Code(simple_rotation_code)
    model_sims = _model.simulations
    if simulations is None:
        new_sim = _model.simulations
    else:
        if isinstance(simulations, str):
            sims = [simulations]
        else:
            sims = simulations
        new_sim = [si for si in model_sims if si.Name in sims]
    for sim in new_sim:
        zone = sim.FindDescendant[Models.Core.Zone]()
        zone.Children.Add(manager)
    man = detect_sowing_managers(_model)
    update = update_manager_code(man.Code, description="Simple Rotation Manager", typeof='Manager', var_name='rotation')
    new_manager = Models.Manager()
    new_manager.Rename('rots')
    new_manager.set_Code(update)
    for sim in new_sim:
        zone = sim.FindDescendant[Models.Core.Zone]()
        zone.Children.Add(new_manager)

    _model.save()
    return _model


def add_method_to_model_tools(method):
    setattr(ModelTools, method.__name__, method)


def add_replacement_folder(simulations):
    rep_old = find_child(simulations, child_class=Models.Core.Folder, child_name='Replacements')
    if not rep_old:
        folder = Models.Core.Folder()
        try:
            folder.Rename('Replacements')
        except AttributeError:
            folder.Name = "Replacements"
        simulations.Children.Add(folder)


def add_model_as_a_replacement(simulations, model_class, model_name):
    if isinstance(model_class, str):
        model_class = validate_model_obj(model_class)
    add_replacement_folder(simulations)
    # ensure that model being added as replacement exists
    if APSIM_VERSION > BASE_RELEASE_NO:
        model_to_replace = find_child(simulations, child_class=model_class, child_name=model_name)
    else:
        model_to_replace = simulations.FindDescendant[model_class](model_name)
    fin_all = find_all_in_scope(simulations, model_class)
    names = ','.join([i.Name for i in fin_all])
    if not model_to_replace:
        raise ValueError(f"model class {model_class} with {model_name} not found in the current simulations. "
                         f"suggested available model names: {names}")
    rep_old = find_child(simulations, child_class=Models.Core.Folder, child_name='Replacements')
    if rep_old:
        # find if the requested model exists in the replacement folder, remove or do nothing
        model_exists = find_child(rep_old, child_class=model_class, child_name=model_name)
        if model_exists:
            rep_old.Children.Remove(model_exists)
        # now we add the found model at the top as replacement
        rep_old.Children.Add(model_to_replace)
    else:
        raise RuntimeError("failed to add replacement folder")
    # check again before exiting caller
    model_exists = find_child(rep_old, child_class=model_class, child_name=model_name)
    if not model_exists:
        raise RuntimeError(f"failed to add model of class{model_class} with identification name: {model_name}")


@dataclass
class ModelTools:
    """
       A utility class providing convenient access to core APSIM model operations and constants.

       Attributes:
           ``ADD`` (callable): Function or class for adding components to an APSIM model.

           ``DELETE`` (callable): Function or class for deleting components from an APSIM model.

           ``MOVE`` (callable): Function or class for moving components within the model structure.

           ``RENAME`` (callable): Function or class for renaming components.

           ``CLONER`` (callable): Utility to clone APSIM models or components.

           ``REPLACE`` (callable): Function to replace components in the model.

           ``MultiThreaded`` (Enum): Enumeration value to specify multi-threaded APSIM runs.

           ``SingleThreaded`` (Enum): Enumeration value to specify single-threaded APSIM runs.

           ``ModelRUNNER`` (class): APSIM run manager that handles simulation execution.

           ``CLASS_MODEL`` (type): The type of the APSIM Clock model, often used for type checks or instantiation.

           ``ACTIONS`` (tuple): Set of supported string actions ('get', 'delete', 'check').

           ``COLLECT`` (callable): Function for forcing memory checks
       """
    ADD = _tools('ADD')
    DELETE = _tools('DELETE')
    MOVE = _tools('MOVE')
    RENAME = _tools('RENAME')
    CLONER = _tools('CLONER')
    REPLACE = _tools('REPLACE')
    MultiThreaded = _tools('MultiThreaded')
    SingleThreaded = _tools('SingleThreaded')
    ModelRUNNER = _tools('ModelRUNNER')
    CLASS_MODEL = _tools('CLASS_MODEL')
    ACTIONS = _tools('ACTIONS')
    COLLECT = _tools('COLLECT')
    String = String
    Double = Double
    Array = Array
    CAST = cast_as
    find_child = find_child
    add_replacement_folder = add_replacement_folder
    find_all_in_scope = find_all_in_scope
    find_child_of_class = find_child_of_class


collect()
# add_method_to_model_tools(find_child)
# add_method_to_model_tools(find_all_in_scope)
# add_method_to_model_tools(add_replacement_folder)

if __name__ == "__main__":
    ...
    from apsimNGpy.core.apsim import ApsimModel

    model = ApsimModel('Maize')
    # ap = configure(model, 'Maize,Soybean', "Soybean", 'fixed',
    #                10, 50, 10, "Bunya", end_date='15-may',start_date= '12-may', fertilize_crop={'Soybean': 30})



import os.path
from pathlib import Path
from typing import Union

import numpy as np
import pandas as pd
from apsimNGpy.core_utils.utils import timer
from apsimNGpy.core.apsim import ApsimModel
from apsimNGpy.core.pythonet_config import Models


class OptimizationBase(ApsimModel):

    def __init__(self, model: Union[Path, dict, str], out_path: Union[str, Path] = None, out: Union[str, Path] = None,
                 lonlat: tuple = None, soil_series: str = 'domtcp', thickness: int = 20, bottomdepth: int = 200,
                 thickness_values: list = None, set_wd=None, **kwargs):

        super().__init__(model, out_path, set_wd, **kwargs)
        self.soil_type = None
        self.SWICON = None
        self.lonlat = lonlat
        self.n_layers = bottomdepth / thickness
        bm = bottomdepth * 10
        if thickness_values is None:
            thickness_values = self.auto_gen_thickness_layers(max_depth=bm, n_layers=int(self.n_layers))
        self.soil_series = soil_series
        self.thickness = thickness
        self.out_path = out_path or out
        self._thickness_values = thickness_values
        self.copy = True
        self.parameters = []
        self.params = []

        if not isinstance(thickness_values, np.ndarray):
            self.thickness_values = np.array(self._thickness_values,
                                             dtype=np.float64)
        else:
            self.thickness_values = self._thickness_values
        if kwargs.get('experiment', False):
            self.create_experiment()

    def check_kwargs(self, path, **kwargs):
        mod_obj = self.Simulations.FindByPath(path)
        if mod_obj is None:
            raise ValueError(f"Could not find model associated with path {path}")
        v_mod = mod_obj.Value
        kas = set(kwargs.keys())

        def _raise_value_error(_path, acceptable, user_info, msg='not a valid attribute'):
            _dif = user_info - acceptable
            if len(_dif) > 0:
                raise ValueError(f"{_dif} is not a valid parameter for {_path}")

        match type(v_mod):
            case Models.Manager:
                kav = {v_mod.Parameters[i].Key for i in range(len(v_mod.Parameters))}
                _raise_value_error(path, kav, kas)

            case Models.Clock:
                acceptable = {'End', "Start"}
                _raise_value_error(path, acceptable, kas)
            case Models.Climate.Weather:
                met_file = kwargs.get('weather_file') or kwargs.get('met_file')
                if met_file is not None:
                    if not os.path.isfile(met_file):
                        raise ValueError(f"{met_file} is not a valid file")
                else:
                    raise ValueError(f"{met_file} file name is needed use key word 'met_file' or 'weather_file'")
            case Models.Surface.SurfaceOrganicMatter:
                accept = {'SurfOM', 'InitialCPR', 'InitialResidueMass',
                          'InitialCNR', 'IncorporatedP', }
                if kwargs == {}:
                    raise ValueError(f"Please supply at least one parameter: value \n '{', '.join(accept)}' for {path}")
                _raise_value_error(path, accept, kas)
            case Models.PMF.Cultivar:
                # Define required parameters
                required_keys = ["commands", "values", "cultivar_manager", "parameter_name", "new_cultivar_name"]
                # Extract input parameters
                param_values = kwargs
                missing = [key for key in required_keys if not param_values.get(key)]

                if missing:
                    raise ValueError(f"Missing required parameter(s): {', '.join(missing)}")

    def add_parameters(self, path: str, *, bounds=None, v_type, **kwargs):
        """
        Add a parameter set associated with a given APSIM model path.

        Parameters
        ----------
        path : str
            The full APSIM component path where the parameters should be edited.
        **kwargs :
            Key-value pairs of parameters to be updated. One and only one value should be
            designated ("?", '', "") to indicate the parameter to be optimized. remember '' is not the same as ''

        Notes
        -----
        - Only one parameter per call should be marked with either "?" or 'fill' (to be filled or edited during calibration).
        - Duplicate entries are ignored.
        @param v_type: var type e.g., int, str, float, bool
        @param bounds: lower and upper limits of x variable
        """
        # Pre-check parameter keys/values for validity in APSIM context
        self.check_kwargs(path, **kwargs)

        # Allow only one parameter to be unspecified as "?" or 'fill'
        missing = [k for k, v in kwargs.items() if v in ("?", 'fill')]
        if len(missing) == 0:
            raise ValueError("At least one parameter value must be left unspecified (e.g., '') for optimization.")
        if len(missing) > 1:
            raise ValueError("Only one parameter value should be unspecified per call.")

        # Check for invalid types (optional)
        for k, v in kwargs.items():
            if not isinstance(k, str):
                raise TypeError(f"Parameter name {k!r} must be a string.")

        # Avoid duplicates
        entry = (path, kwargs)
        if entry not in self.parameters:
            self.parameters.append(entry)
        args = (path, bounds, v_type, kwargs)
        out_args = dict(zip(['path', "v_type", 'bounds', 'kwargs'], args))
        if out_args not in self.params:
            self.params.append(out_args)

    def set_param(self, path, x_value):
        st = path.split('.')
        param_name = st[-1]
        base_path = st[:-1]
        base_path = '.'.join(base_path)
        self.edit_model_by_path(base_path, **{param_name: x_value})

    @timer
    def _do_model_edit(self, x):
        """Objective function: apply `x` values to APSIM model during parameter optimization or calibration."""
        if not self.parameters:
            raise ValueError("No parameters defined for calibration.")

        if len(x) != len(self.parameters):
            raise ValueError(f"Expected {len(self.parameters)} parameters' values, but got {len(x)}.")

        for i, value in enumerate(x):
            path, param_dict = self.parameters[i]
            updated_params = {k: value for k in param_dict.keys() if param_dict[k] in ('?', "")}
            pp = param_dict | updated_params  # Python 3.9+: right-hand overrides left-hand

            self.edit_model_by_path(path=path, **pp)

    @staticmethod
    def is_incomplete_date(series: pd.Series) -> bool:
        """
        Determine if a pandas datetime series was likely created from year, month or days only values.
        Assumes the default day and month values are 1 (i.e., January 1st).
        """
        if not pd.api.types.is_datetime64_any_dtype(series):
            raise ValueError("Input series must be datetime type")

        # Check if all dates are January 1st
        return (series.dt.month == 1).all() and (series.dt.day == 1).all()

    @staticmethod
    def _match(text):
        import re
        pattern = r'\[[^\]]+\]\.\w+'
        return re.findall(pattern, text)

    @property
    def n_vars(self):
        return len(self.parameters)

    def clean_data(self, obs: Union[pd.DataFrame, str, Path], *,
                   data_table: str,
                   output_column: str,
                   time_step: str = 'Year',
                   obs_time_column: str = 'year'):
        """
        Align observed data with APSIM simulation results.

        Parameters
        ----------
        obs : Union[pd.DataFrame, str, Path]
            The observed data, either as a DataFrame or a file path to a CSV.
        data_table : str
            The name of the APSIM report table to extract simulation results from.
        output_column : str
            The output column of interest from the APSIM results.
        time_step : str, default 'Year'
            One of {'Year', 'Month', 'Day'} or full date.
        obs_time_column : str, default 'year'
            The time column in the observed data.

        Returns
        -------
        pd.DataFrame
            A DataFrame with columns: [date, time_step, apsim, observed]
        """
        DATE = 'date'

        # Accept a file path or DataFrame
        match obs:
            case pd.DataFrame():
                obs = obs.copy(deep=True)
            case str() | Path():
                obs = pd.read_csv(str(obs))
            case _:
                raise TypeError(f"Unsupported type for 'obs': {type(obs)}")

        if obs_time_column not in obs.columns:
            raise KeyError(f"Column '{obs_time_column}' not found in observed data")

        obs_date = pd.to_datetime(obs[obs_time_column])

        def _proc_df(_obs, _data_table, column, _time_step, _obs_time_column):
            time = _time_step.capitalize()
            if time not in {'Year', 'Month', 'Day'}:
                raise ValueError("`time_step` must be one of {'Year', 'Month', 'Day'}")

            # Clock variable to add to APSIM
            if self.is_incomplete_date(obs_date):
                clock_variable = f'[Clock].Today.{time} as {_obs_time_column}'
            else:
                clock_variable = f'[Clock].Today as {_obs_time_column}'

            # Validate report existence
            available_reports = self.inspect_model('Models.Report', fullpath=False)
            if not isinstance(_data_table, str):
                raise TypeError("Data table name must be a string")
            if not available_reports:
                raise ValueError("No report tables found in the APSIM model")
            if _data_table not in available_reports:
                suggestion = 'Did you mean' if len(available_reports) == 1 else 'Did you mean any of'
                raise ValueError(
                    f"Data table '{_data_table}' not found. {suggestion} '{', '.join(available_reports)}'?"
                )

            # Inject Clock variable to report
            self.add_report_variable(
                [clock_variable, f'[Clock].Today as {DATE}'],
                _data_table
            )

            dtf = self.run(_data_table).results

            if column not in dtf.columns:
                raise ValueError(f"Column '{column}' not found in simulation output table.")

            obs_time = _obs[_obs_time_column]
            sdf = dtf[dtf[_obs_time_column].isin(obs_time)]

            if sdf.shape[0] != _obs.shape[0]:
                raise ValueError(
                    f"Mismatch in row count:\nSimulated: {sdf.shape[0]}\nObserved: {_obs.shape[0]}"
                )

            _obs.sort_values(by=_obs_time_column, inplace=True)

            result = pd.DataFrame()
            if DATE != _obs_time_column and self.is_incomplete_date(obs_date):
                result[DATE] = sdf[DATE]
            elif self.is_incomplete_date(obs_date):
                result[DATE] = _obs[_obs_time_column]

            result[_time_step] = sdf[_obs_time_column]
            result['apsim'] = sdf[output_column]
            result['observed'] = _obs[output_column]

            return result

        return _proc_df(obs, _data_table=data_table, column=output_column,
                        _time_step=time_step, _obs_time_column=obs_time_column)



import numpy as np
import pytest
from apsimNGpy.optimizer.sampler import float_random_sampling, integer_random_sampling, boolean_random_sampling


def test_float_random_sampling_default_bounds():
    n_var, n = 3, 5
    samples = float_random_sampling(n_var, n)
    assert samples.shape == (n, n_var)
    assert np.all(samples >= 0.0) and np.all(samples < 1.0)


def test_float_random_sampling_custom_bounds():
    n_var, n = 2, 10
    xl = np.array([10, 20])
    xu = np.array([20, 30])
    samples = float_random_sampling(n_var, n, bounds=(xl, xu))
    assert samples.shape == (n, n_var)
    assert np.all(samples >= xl) and np.all(samples <= xu)


def test_float_random_sampling_invalid_bounds_shape():
    with pytest.raises(ValueError):
        float_random_sampling(3, 5, bounds=(np.array([1, 2]), np.array([3, 4])))


def test_float_random_sampling_invalid_bounds_order():
    with pytest.raises(ValueError):
        float_random_sampling(2, 5, bounds=(np.array([5, 10]), np.array([4, 9])))


def test_integer_random_sampling_shape_and_range():
    n_var, n = 2, 50
    bounds = ((10, 20), (100, 200))
    samples = integer_random_sampling(n_var, n, bounds)
    assert samples.shape == (n, n_var)
    for i in range(n_var):
        assert np.all(samples[:, i] >= bounds[0][i])
        assert np.all(samples[:, i] <= bounds[1][i])


def test_boolean_random_sampling_shape_and_values():
    n_var, n = 4, 10
    samples = boolean_random_sampling(n_var, n)
    assert samples.shape == (n, n_var)
    assert samples.dtype == bool
    assert np.all(np.logical_or(samples == True, samples == False))


if __name__ == "__main__":
        ...
        import sys

        pytest.main([__file__])



"""
This module offers a procedural alternative other than object-oriented approach provided in api and ApsimModel classes
"""
import os
import uuid
from typing import Union

from apsimNGpy.core import pythonet_config
from apsimNGpy.core_utils.utils import timer
from apsimNGpy.exceptions import NodeNotFoundError

pyth = pythonet_config
# now we can safely import C# libraries
from System.Collections.Generic import *
from System import *
import Models

import json
from os.path import (realpath)
import shutil
from pathlib import Path
from apsimNGpy.core.config import get_apsim_bin_path, apsim_version, load_crop_from_disk
import subprocess
from dataclasses import dataclass
from typing import Any
from pathlib import Path
from apsimNGpy.core.cs_resources import CastHelper as CastHelpers
from apsimNGpy.core.pythonet_config import get_apsim_file_reader, get_apsim_file_writer
from apsimNGpy.core.pythonet_config import is_file_format_modified

GLOBAL_IS_FILE_MODIFIED = is_file_format_modified()
scratch_dir = Path.cwd().joinpath('scratch')
scratch_dir.mkdir(exist_ok=True)
SCRATCH = os.environ.get('WS', str(scratch_dir))


def stamp_name_with_version(file_name):
    """
    we stamp every file name with the version, which allows the user to open it in the appropriate version it was created
    @param file_name: path to the would be.apsimx file
    @return: path to the stamped file
    """
    _version = apsim_version()
    destination = Path(file_name).resolve()
    dest_path = destination.with_name(
        destination.name.replace(".apsimx", f"{_version}.apsimx")
    )
    return dest_path


def to_model_from_string(json_string, file_name):
    loader = get_apsim_file_reader()
    return loader[Models.Core.Simulations](json_string, None, True, fileName=file_name)


def to_json_string(_model: Models.Core.Simulation):
    """We first determine whether the model is loaded from APSIM>CORE."""

    n_model = getattr(_model, 'Node', _model)
    if hasattr(n_model, 'ToJSONString'):
        return n_model.ToJSONString()

    writer = get_apsim_file_writer()
    if _model:
        return writer(_model)
    raise ValueError(f'failed to convert model:`{_model}` to strings ')


@dataclass
class ModelData:
    """
    This is a meta-data container for the loaded models
    """
    IModel: Models
    path: str
    datastore: Any = None
    DataStore: Any = None
    results: Any = None
    met_path: str = ""
    Node: Any = None
    Simulations: Any = None


def get_model(obj):
    out_model = getattr(obj, 'Model', obj)  # for new versions
    out_model = getattr(out_model, 'NewModel', out_model)  # previous versions
    return out_model


def load_from_dict(dict_data, out):
    str_ = json.dumps(dict_data)
    out = realpath(out)
    return to_model_from_string(str_, file_name=out)


version = apsim_version()


def copy_file(
        source: Union[str, Path],
        destination: Union[str, Path] = None,
        wd: Union[str, Path] = None):
    """
    Copy an APSIMX file to a new location with an updated version tag.

    Args:
        source (str | Path): Path to the source APSIMX file.

        destination (str | Path, optional): Path to the destination APSIMX file.
            If None, a temporary file will be created in `wd`.

        wd (str | Path, optional): Working directory for temporary file creation.
            Defaults to SCRATCH.

    Returns:
        Path: Path to the copied file.
    """
    # Ensure working directory is set. defaults to current working directory
    wd = Path(wd or SCRATCH)

    # Determine destination path
    if destination:
        dest_path = stamp_name_with_version(destination)
    else:
        dest_path = wd / f"temp_{uuid.uuid1()}_{version}.apsimx"

    # Perform copy
    shutil.copy2(source, dest_path)
    return str(dest_path)


def save_model_to_file(_model, out=None):
    """Save the model

        Parameters
        ----------

            Returns the filename or the specified out name
        """
    # Determine the output path
    _model = get_model(_model)

    final_out_path = out or '_saved_model.apsimx'

    json_string = to_json_string(_model)
    # Serialize the model to JSON string

    # Save the JSON string to the determined output path
    with open(final_out_path, "w", encoding='utf-8') as f:
        f.write(json_string)
    return final_out_path


def covert_to_model(object_to_convert):
    return getattr(object_to_convert, 'NewModel', object_to_convert)


def load_from_path(path2file, method='string'):
    """

    path2file: path to apsimx file.

    method: str with string, we direct the method to first convert the file.
    into a string using json and then use the APSIM in-built method to load the file with file, we read directly from
    the file path. This is slower than the former.
    """

    f_name = realpath(path2file)
    with open(f_name, "r+", encoding='utf-8') as apsimx:
        app_ap = json.load(apsimx)
    string_name = json.dumps(app_ap)

    method = method.lower()

    loader = get_apsim_file_reader(method)

    match method:
        case 'string':
            __model = loader[Models.Core.Simulations](string_name, None, True, fileName=f_name)
            __model = getattr(__model, 'NewModel', __model)

        case 'file':
            __model = loader[Models.Core.Simulations](f_name, None, True)
            __model = getattr(__model, 'NewModel', __model)
        case _:
            raise NotImplementedError('Unsupported method for reading apsim json file')

    new_model = covert_to_model(__model)

    return new_model


def load_apsim_model(model=None, out_path=None, file_load_method='string', met_file=None, wd=None, **kwargs):
    """
    Load an APSIMX model from a file path, dictionary, or in-memory object.

    Parameters:
        model (str | dict | Simulation | None): The APSIM model source. Can be a file path, dictionary,
                                                APSIM simulation object, or None.
        out_path (str | Path, optional): The output path to save the working copy of the model.
                                         If None, a new filename will be generated.
        file_load_method (str): How to load the file (e.g., 'string', 'json').
        met_file (str, optional): Path to the associated meteorological file.
        wd (str, optional): Working directory for temporary file operations.
        **kwargs: Additional options (reserved for future use).

    Returns:
        {ModelData}: A dataclass container with paths, model object, and metadata.
    """
    if isinstance(model, Path):
        model = str(model)

    out = {}  # Store final output path

    match model:
        case dict():
            output = out_path or f"{uuid.uuid1()}.apsimx"
            out['path'] = output
            Model = load_from_dict(model, output)

        case str():
            if not model.endswith('.apsimx'):
                model = load_crop_from_disk(crop=model, out=out_path, work_space=wd)

            copy_to = copy_file(model, destination=out_path, wd=wd)

            out['path'] = copy_to
            Model = load_from_path(copy_to, file_load_method)

        case None:
            raise ValueError("Model cannot be None")

        case _:
            raise NotImplementedError(f"Unsupported model type: {type(model)}")

    _Model = get_model(Model)
    node = _Model

    if GLOBAL_IS_FILE_MODIFIED:
        g_model = getattr(_Model, 'Model', _Model)
        out_model = CastHelpers.CastAs[Models.Core.Simulations](g_model)
    else:
        out_model = _Model

    _out_path = out.get('path')

    datastore_path = str(Path(_out_path).with_suffix('.db')) if _out_path else ""

    if hasattr(out_model, "FindChild"):
        DataStore = out_model.FindChild[Models.Storage.DataStore]()
    else:
        DataStore = [i for i in out_model.Children if i.Name == 'DataStore']
        if DataStore:
            DataStore = DataStore[0]

        else:
            DataStore = ''
    if DataStore:
        DataStore = CastHelpers.CastAs[Models.Storage.DataStore](DataStore)
    return ModelData(
        IModel=out_model,
        path=str(_out_path),
        datastore=datastore_path,
        DataStore=DataStore,
        results=None,
        met_path=met_file,
        Node=node,
        Simulations=Model
    )


def load_as_dict(file: Models.Core.ApsimFile) -> dict:
    """loads apsimx model from a pathlib"""
    model_simulations = load_from_path(file)
    json_string = Models.Core.ApsimFile.FileFormat.WriteToString(model_simulations)
    return json.loads(json_string)


def recompile(_model, out=None, met_path=None, ):
    """ recompile without saving to disk useful for recombining the same model on the go after updating management scripts

            Parameters
            ----------
            out : str, optional path to save the model to

                :param met_path: path to met file
                :param out: out path name for database reconfiguration
                :param _model:APSIM Models.Core.Simulations object
                returns named tuple with a recompiled model
            """
    # Determine the output path

    final_out_path = out or _model.path

    # Serialize the model to JSON string

    json_string = to_json_string(_model.Simulations)
    if GLOBAL_IS_FILE_MODIFIED:
        model_node = getattr(_model.Simulations, 'Node', _model.Simulations)
        json_string = model_node.ToJSONString()
    Model = to_model_from_string(json_string, file_name=_model.path)

    _Model = False

    _Model = get_model(Model)
    if GLOBAL_IS_FILE_MODIFIED:
        out_model = CastHelpers.CastAs[Models.Core.Simulations](_Model)
    else:
        out_model = _Model

    datastore = str(Path(_model.path).with_suffix('.db'))

    try:
        DataStore = out_model.FindChild[Models.Storage.DataStore]()
    except AttributeError:
        DataStore = [i for i in out_model.Children if i.Name == 'DataStore']
        if DataStore:
            DataStore = DataStore[0]
        else:
            DataStore = ''
    if DataStore:
        DataStore = CastHelpers.CastAs[Models.Storage.DataStore](DataStore)
    # need to make ModelData a constant and named outside the script for consistency across scripts
    # ModelData = namedtuple('model_data', ['IModel', 'path', 'datastore', "DataStore", 'results', 'met_path'])
    return ModelData(IModel=out_model,
                     path=final_out_path,
                     datastore=datastore,
                     DataStore=DataStore,
                     results=None,
                     met_path=met_path,
                     Node=_Model,
                     Simulations=Model)


def model_from_string(mod, out=None):
    """
    try to load a model from string, and it is not converted to the corresponding reference type
    @param out: out name or path of the model
    @param mod:
    @return:
    """
    if str(mod).endswith('.apsimx'):
        path2file = mod
    else:
        out = out or os.path.realpath(f'{mod}_testformatx.apsimx')
        path2file = load_crop_from_disk(mod, out=out)

    f_name = realpath(path2file)

    with open(f_name, "r+", encoding='utf-8') as apsimx:
        app_ap = json.load(apsimx)
        string_name = json.dumps(app_ap)
        loader = get_apsim_file_reader(method='string')

        mod = loader[Models.Core.Simulations](string_name, None, initInBackground=True)
        mod = getattr(mod, "Model", mod)

        mod = covert_to_model(mod)

        return mod


@timer
def run_model_externally(model):
    # Define the APSIM executable path (adjust if needed)
    apsim_exe = Path(get_apsim_bin_path()) / 'Models.exe'

    # Define the APSIMX file path
    apsim_file = model

    # Run APSIM with the specified file
    try:
        result = subprocess.run([apsim_exe, apsim_file])

        print("APSIM Run Successful!")

        # df = read_db_table(datastore, table)
        # return df
    except subprocess.CalledProcessError as e:
        print("Error running APSIM:")
        print(e.stderr)  # Print APSIM error message if execution fails


def get_node_by_name(node, name):
    if hasattr(node, 'Node'):
        node = node.Node
    if hasattr(node, 'Walk'):
        for n in node.Walk():
            if n.Name == name:
                return n
    else:
        raise AttributeError(f"Node supplied has no attribute: Walk")
    raise NodeNotFoundError(f'Node with supplied name: `{name}` was not found')
    #


def get_node_by_path(node, node_path):
    """
    get a node by path
    @param node: node object or APSIM.Core object
    @param node_path: node path
    @return: node object if found. raise NodeNotFoundError
    """

    if hasattr(node, 'Node'):
        node = node.Node
    if hasattr(node, 'Walk'):
        for n in node.Walk():

            if n.get_FullNameAndPath() == node_path:
                return n
    else:
        raise AttributeError(f"Node supplied has no attribute: Walk")
    raise NodeNotFoundError(f'Node with supplied path: `{node_path}` was not found')


def get_node_string(node):
    """

    @param node: node object
    @return: a string name for the node e.g., Models.Clock
    """
    return node.ToString()


def get_attributes(obj):
    attributes = set()
    d = dir(obj)
    for i in d:
        if not i.startswith('__'):
            attributes.add(i)
    return attributes


if __name__ == '__main__':
    pat = load_crop_from_disk('Maize')
    load = load_apsim_model('Maize')
    print(load.DataStore, 'datastore')
    p, model, model2 = load.Node, load.IModel, load.IModel

    getattr(Models.Core.ApsimFile, "FileFormat", None)
    # set_apsim_bin_path(r'/Applications/APSIM2025.2.7670.0.app/Contents/Resources/bin')
    to_json_string(model2)



import numpy as np
import matplotlib.pyplot as plt
from scipy.optimize import minimize
from pymoo.core.problem import ElementwiseProblem
from pymoo.algorithms.moo.nsga2 import NSGA2
from pymoo.optimize import minimize as pymoo_minimize
from pymoo.operators.sampling.rnd import FloatRandomSampling, IntegerRandomSampling
from pymoo.termination import get_termination

# Objective function
def fun(x):
    return x**2

# --- Nelder-Mead ---
res_nm = minimize(fun, x0=[-10], method='powell',bounds=[(-10, 10)], options={'maxiter': 1000})
print(res_nm)
# --- NSGA-II on single-objective ---
class MyProblem(ElementwiseProblem):
    def __init__(self):
        super().__init__(n_var=1, n_obj=1, n_constr=0, xl=-10.0, xu=10.0)

    def _evaluate(self, x, out, *args, **kwargs):
        out["F"] = fun(x)

problem = MyProblem()
algorithm = NSGA2(
    pop_size=10,
    sampling= IntegerRandomSampling(),
    eliminate_duplicates=True,

)

termination = get_termination("n_gen", 1000)

res_nsga2 = pymoo_minimize(problem, algorithm, termination, seed=1, verbose=False)
print(res_nsga2.X,'\n')
nsg2 = res_nsga2.F
if nsg2 < res_nm.fun:
    print('NSG2 has won')
elif nsg2 == res_nm.fun:
    print('draw')
else:
    print('NSG2 has lost', '\n')


print(res_nsga2.F)




if __name__ == '__main__':
    from apsimNGpy.optimizer.single import  ContinuousVariable, MixedVariable
    from apsimNGpy.core.apsim import ApsimModel
    maize_model = "Maize"
    obs = [
        7000.0,
        5000.505,
        1000.047,
        3504.000,
        7820.075,
        7000.517,
        3587.101,
        4000.152,
        8379.435,
        4000.301
    ]

    print('testing continous variable')
    # there are two ways to initialise a problem from both classes

    # 1. inherit from the classes as follows
    class Problem(ContinuousVariable):
        def __init__(self, apsim_model: 'ApsimNGpy.Core.Model', obs, max_cache_size=400, objectives=None,
                     decision_vars=None):
            super().__init__(apsim_model=apsim_model, max_cache_size=max_cache_size, objectives=objectives,
                             decision_vars=decision_vars)
            self.obs = obs

        def evaluate_objectives(self, **kwargs):
            # set up everything you need here. This will act as the objective function
            # tip run the apsim model here, get the results and decide on how to evaluate the objective
            predicted = self.apsim_model.run(verbose=False).results.Yield

            ans = self.rmse(self.obs, predicted)

            return -predicted.mean()

    # now intitialize the problem

    maize_model = ApsimModel('Maize')
    problem = Problem(maize_model, obs, objectives=None, decision_vars=None)

    # approach 2.
    # 2. define the objective and supply it directly. what is needed is to access the pandas data and decide on how to evaluate the objective function
    #define objective function, should take in a pandas data frame querry one the columns from the predicted table and return the evaluation results
    def maximize_yield(df):
        ans = -df.Yield.mean()
        return ans
    # Adding control variables
    # There are also two ways to add control variables, a. ka decision variables. using add control or supplying the decision variable directly during initialisation. we will get to the later later on, but
    # first let use add_control vars
    problem.add_control(
        **{'path': '.Simulations.Simulation.Field.Fertilise at sowing', 'Amount': "?", "bounds": [50, 300],
           "v_type": "qrandint"}, start_value=150)
    problem.add_control(
        **{'path': '.Simulations.Simulation.Field.Sow using a variable rule', 'Population': "?", 'v_type': 'int',
           'bounds': [4, 14]}, start_value=8)

    # ready to optimize you variables. no need to supply starting values, bounds etc as those are defined with the add_control method but you can concentrate on supplying other arguments such a smethod to use. the method to use depends on the nature of the problem

    resc = problem.minimize_with_alocal_solver( method  ='Powell',  options={
        # 'xatol': 1e-4,      # absolute error in xopt between iterations
        # 'fatol': 1e-4,      # absolute error in func(xopt) between iterations
        'maxiter': 100,    # maximum number of iterations
        'disp': True ,      # display optimization messages

    })

    depc = problem.minimize_with_de(popsize=10, maxiter=100, polish=False)

    print('testing mixed variable')
    # mixed problem takes in all types of variables from discrete to continnous and also provide some flexibility in guiding the sampling processing, which help to speed up the optimization pricess
    maize_model = ApsimModel('Maize')
    problem = MixedVariable(maize_model, objectives=maximize_yield)
    problem.add_control(
        **{'path': '.Simulations.Simulation.Field.Fertilise at sowing', 'Amount': "?", "bounds": None,
           "v_type": "choice"}, start_value=150, categories=[100,150, 200, 250, 300])
    problem.add_control(
        **{'path': '.Simulations.Simulation.Field.Sow using a variable rule', 'Population': "?", 'v_type': 'qrandint',
           'bounds': [4, 14]}, start_value=8, q=2, )
    # de_res = problem.optimize_with_differential_evolution()
    res_m = problem.minimize_with_alocal_solver(method='Powell', )
    de_m = problem.minimize_with_de(popsize=20, polish=True)
